# @file conf_mgmt.py
# Handle Edk2 Conf management
# Customized for edk2-pytools-extensions based build and supports dynamic Visual studio support 2017++
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Handle Edk2 Conf Management.

Customized for edk2-pytool-extensions based build and support dynamic Visual studio support 2017++
"""
import logging
import os
import shutil
import time

from edk2toolext.environment import version_aggregator


class ConfMgmt():
    """Handles Edk2 Conf Management."""
    def __init__(self):
        """Init an empty ConfMgmt object."""
        self.Logger = logging.getLogger("ConfMgmt")
        self.delay_time_in_seconds = 30

    def _set_delay_time(self, time_in_seconds):
        """Allow changing the warning time for out of date templates."""
        self.delay_time_in_seconds = time_in_seconds

    def populate_conf_dir(self, conf_folder_path: str, override_conf: bool, conf_template_source_list: list) -> None:
        """Compare the conf dir files to the template files.

        Copy files if they are not present in the conf dir or the override parameter is set.

        Args:
            conf_folder_path (str): folder path to output conf location (absolute path)
            override_conf (bool):  boolean to indicate if templates files should replace conf files
                regardless of existence or version.
            conf_template_source_list (list): priority list of folder path that might contain a "Conf"
                older with template files to use

        Note: When complete the conf_folder_path dir must be setup for edk2 builds
        """
        #  make folder to conf path if needed
        os.makedirs(conf_folder_path, exist_ok=True)

        files = ["target.txt", "tools_def.txt", "build_rule.txt"]  # add more files here

        # make output conf files list based on files
        outfiles = [os.path.join(conf_folder_path, f) for f in files]

        # make template list based on files
        templatefiles = [os.path.join("Conf", os.path.splitext(f)[0] + ".template") for f in files]

        # loop thru each Conf file needed
        for x in range(0, len(outfiles)):
            template_file_path = None

            # find template file given multiple root locations
            for r in conf_template_source_list:
                p = os.path.join(r, templatefiles[x])
                if os.path.isfile(p):
                    template_file_path = p
                    break

            if (template_file_path is None):
                self.Logger.critical(
                    "Failed to find Template file for %s" % outfiles[x])
                raise Exception("Template File Missing", outfiles[x])
            else:
                self.Logger.debug(f"Conf file template: {template_file_path}")

            # have template now - now copy if needed
            self._copy_conf_file_if_necessary(outfiles[x], template_file_path, override_conf)

            # Log Version for reporting
            version_aggregator.GetVersionAggregator().ReportVersion(outfiles[x], self._get_version(outfiles[x]),
                                                                    version_aggregator.VersionTypes.INFO)

    def _get_version(self, conf_file: str) -> str:
        """Parse the version from the conf_file.

        version should be in #!VERSION={value} format

        !!! note
            "0.0" is returned if no version is found
        """
        version = "0.0"
        with open(conf_file, "r") as f:
            for line in f.readlines():
                if (line.startswith("#!VERSION=")):
                    try:
                        version = str(float(line.split("=")[1].split()[0].strip()))
                        break
                    except Exception:
                        pass
        return version

    def _is_older_version(self, conf_file: str, template_file: str) -> bool:
        """Determines whether the conf or template file is older.

        Given a conf_file and a template_file file determine if
        the conf_file has an older version than the template.

        Args:
            conf_file (str): path to current conf_file
            template_file (str): path to template file that is basis
                for the the conf_file.
        """
        conf = 0
        template = 0

        try:
            conf = float(self._get_version(conf_file))
            logging.debug("Conf version: %s", str(conf))
            template = float(self._get_version(template_file))
            logging.debug("Template Version: %s", str(template))
        except Exception:
            logging.error("Failed to get version from file")
        finally:
            return (conf < template)

    def _copy_conf_file_if_necessary(self, conf_file: str, template_file: str, override_conf: bool) -> None:
        """Copy template_file to conf_file if policy applies.

        Args:
            conf_file (str): path to final conf file location
            template_file (str): template file to copy if policy applies
            override_conf (bool): flag indication to override regardless of policy
        """
        if not os.path.isfile(conf_file):
            # file doesn't exist.  copy template
            self.Logger.debug(f"{conf_file} file not found.  Creating from Template file {template_file}")
            shutil.copy2(template_file, conf_file)

        elif (override_conf):
            # caller requested override even for existing file
            self.Logger.debug(f"{conf_file} file replaced as requested")
            shutil.copy2(template_file, conf_file)

        else:
            # Both file exists.  Do a quick version check
            if (self._is_older_version(conf_file, template_file)):
                # Conf dir file is older.  Warn user.
                self.Logger.critical(f"{conf_file} file is out-of-date.  Please update your conf files!")
                self.Logger.critical("Sleeping 30 seconds to encourage update....")
                time.sleep(self.delay_time_in_seconds)
            else:
                self.Logger.debug(f"Conf file {conf_file} up-to-date")



# @file EnvironmentDescriptorFiles.py
# This module contains code for working with the JSON environment
# descriptor files. It can parse the files, validate them, and return
# objects representing their contents.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This module contains code for working with the JSON environment descriptor files.

It can parse the files, validate them, and return objects representing their contents.
"""
import os

import yaml


class PathEnv(object):
    """Path env object that is created from the descriptor file.

    Attributes:
        scope (string): scope the path env is associated with
        flags (List[str]): flags associated with the path env
        var_name (string): ENV var to set with the object
        descriptor_location (string): location of the PathEnv
        published_path (string): location of the PathEnv
    """
    def __init__(self, descriptor):
        """Init with the descriptor information."""
        super(PathEnv, self).__init__()

        #
        # Set the data for this object.
        #
        self.scope = descriptor['scope']
        self.flags = descriptor['flags']
        self.var_name = descriptor.get('var_name', None)

        self.descriptor_location = os.path.dirname(
            descriptor['descriptor_file'])
        self.published_path = self.descriptor_location


class DescriptorFile(object):
    """The base class for the different types of descriptor files.

    Attributes:
        file_path (str): descriptor file path
        descriptor_contents (Dict): Contents of the descriptor file
    """
    def __init__(self, file_path):
        """Loads the contents of the descriptor file and validates.

        Args:
            file_path (str): path to descriptor file

        Raises:
            (ValueError): Missing specified value from descriptor file
        """
        super(DescriptorFile, self).__init__()

        self.file_path = file_path
        self.descriptor_contents = None

        with open(file_path, 'r') as file:
            try:
                self.descriptor_contents = yaml.safe_load(file)
            except Exception:
                pass  # We'll pick up this error when looking at the data.

        #
        # Make sure that we loaded the file successfully.
        #
        if self.descriptor_contents is None:
            raise ValueError(
                "Could not load contents of descriptor file '%s'!" % file_path)

        # The file path is an implicit descriptor field.
        self.descriptor_contents['descriptor_file'] = self.file_path

        # All files require a scope.
        if 'scope' not in self.descriptor_contents:
            raise ValueError("File '%s' missing required field '%s'!" %
                             (self.file_path, 'scope'))

        # If a file has flags, make sure they're sane.
        if 'flags' in self.descriptor_contents:
            # If a flag requires a name, make sure a name is provided.
            for name_required in ('set_shell_var', 'set_build_var'):
                if name_required in self.descriptor_contents['flags']:
                    if 'var_name' not in self.descriptor_contents:
                        raise ValueError(
                            "File '%s' has a flag requesting a var, but does not provide 'var_name'!" % self.file_path)

        # clean up each string item for more reliable processing
        for (k, v) in self.descriptor_contents.items():
            if (isinstance(v, str)):
                self.descriptor_contents[k] = self.sanitize_string(v)

    def sanitize_string(self, s):
        """Clean up a string "value" in the descriptor file."""
        # Perform any actions needed to clean the string.
        return s.strip()


class PathEnvDescriptor(DescriptorFile):
    """Descriptor File for a PATH ENV."""
    def __init__(self, file_path):
        """Inits the descriptor as a PathEnvDescriptor from the provided path.

        Loads the contents of the filepath into descriptor_contents

        Args:
            file_path (str): path to the yaml descriptor file
        """
        super(PathEnvDescriptor, self).__init__(file_path)

        #
        # Validate file contents.
        #
        # Make sure that the required fields are present.
        for required_field in ('flags',):
            if required_field not in self.descriptor_contents:
                raise ValueError("File '%s' missing required field '%s'!" % (
                    self.file_path, required_field))


class ExternDepDescriptor(DescriptorFile):
    """Descriptor File for a External Dependency.

    Attributes:
            descriptor_contents (Dict): Contents of the Descriptor yaml file
            file_path (PathLike): path to the descriptor file
    """
    def __init__(self, file_path):
        """Inits the descriptor as a ExternDepDescriptor from the provided path.

        Loads the contents of the filepath into descriptor_contents

        Args:
            file_path (str): path to the yaml descriptor file
        """
        super(ExternDepDescriptor, self).__init__(file_path)

        #
        # Validate file contents.
        #
        # Make sure that the required fields are present.
        for required_field in ('scope', 'type', 'name', 'source', 'version'):
            if required_field not in self.descriptor_contents:
                raise ValueError("File '%s' missing required field '%s'!" % (
                    self.file_path, required_field))


class PluginDescriptor(DescriptorFile):
    """Descriptor File for a Plugin.

    Attributes:
        descriptor_contents (Dict): Contents of the Descriptor yaml file
        file_path (PathLike): path to the descriptor file
    """
    def __init__(self, file_path):
        """Inits the descriptor as a PluginDescriptor from the provided path.

        Loads the contents of the filepath into descriptor_contents

        Args:
            file_path (str): path to the yaml descriptor file
        """
        super(PluginDescriptor, self).__init__(file_path)

        #
        # Validate file contents.
        #
        # Make sure that the required fields are present.
        for required_field in ('scope', 'name', 'module'):
            if required_field not in self.descriptor_contents:
                raise ValueError("File '%s' missing required field '%s'!" % (
                    self.file_path, required_field))

        # Make sure the module item doesn't have .py on the end
        if (self.descriptor_contents["module"].lower().endswith(".py")):
            # remove last 3 chars
            self.descriptor_contents["module"] = self.descriptor_contents["module"][:-3]



# @file uefi_build.py
# This module contains code that supports the Tianocore Edk2 build system
# This class is designed to be subclassed by a platform to allow
# more extensive and custom behavior.
##
# Copyright (c) Microsoft Corporation
# Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code that supports the Tianocore Edk2 build system.

This class is designed to be subclassed by a platform to allow more extensive
and custom behavior.
"""

import datetime
import logging
import os
import time
import traceback
from collections import namedtuple

from edk2toollib.uefi.edk2.parsers.dsc_parser import DscParser
from edk2toollib.uefi.edk2.parsers.fdf_parser import FdfParser
from edk2toollib.uefi.edk2.parsers.targettxt_parser import TargetTxtParser
from edk2toollib.uefi.edk2.path_utilities import Edk2Path
from edk2toollib.utility_functions import RemoveTree, RunCmd

from edk2toolext import edk2_logging
from edk2toolext.environment import conf_mgmt, shell_environment
from edk2toolext.environment.multiple_workspace import MultipleWorkspace
from edk2toolext.environment.plugintypes.uefi_build_plugin import IUefiBuildPlugin


class UefiBuilder(object):
    """Object responsible for the full build process.

    The following steps are completed by the `UefiBuilder` and is overridable
    by the platform:

    1. `PlatformPreBuild()`
    2. `UefiBuildPlugins` that implement `do_pre_build()`
    3. `Build()` (should not be overridden)
    4. `UefiBuildPlugins` that implement `do_post_build()`
    5. `PlatformFlashImage()`

    Attributes:
        SkipPreBuild (bool): Skip Pre Build or not
        SkipPostBuild (bool): Skip Post Build or not
        SkipBuild (bool): Skip Build or not
        FlashImage (bool): Flash the image not
        Clean (bool): Clean the build directory or not
        Update Conf (bool): Update the conf or not
        env (VarDict): Special dictionary containing build and env vars
        mws (MultipleWorkspace): DEPRECATED. Use self.edk2path
        edk2path (Edk2Path): path utilities for manipulating edk2 paths, packages, and modules
        ws (str): Workspace root dir
        pp (str): packagespath separated by os.pathsep
        Helper (HelperFunctions): object containing registered helper functions
        pm (PluginManager): The plugin manager
    """

    def __init__(self):
        """Inits an empty UefiBuilder."""
        self.SkipBuild = False
        self.SkipPreBuild = False
        self.SkipPostBuild = False
        self.FlashImage = False
        self.ShowHelpOnly = False
        self.OutputBuildEnvBeforeBuildToFile = None
        self.Clean = False
        self.UpdateConf = False
        self.OutputConfig = None

    def AddPlatformCommandLineOptions(self, parserObj):
        """Adds command line options to the argparser.

        Args:
            parserObj (argparser): argparser object
        """
        parserObj.add_argument("--SKIPBUILD", "--skipbuild", "--SkipBuild", dest="SKIPBUILD",
                               action='store_true', default=False, help="Skip the build process")
        parserObj.add_argument("--SKIPPREBUILD", "--skipprebuild", "--SkipPrebuild", dest="SKIPPREBUILD",
                               action='store_true', default=False, help="Skip prebuild process")
        parserObj.add_argument("--SKIPPOSTBUILD", "--skippostbuild", "--SkipPostBuild", dest="SKIPPOSTBUILD",
                               action='store_true', default=False, help="Skip postbuild process")
        parserObj.add_argument("--FLASHONLY", "--flashonly", "--FlashOnly", dest="FLASHONLY",
                               action='store_true', default=False, help="Flash rom after build.")
        parserObj.add_argument("--FLASHROM", "--flashrom", "--FlashRom", dest="FLASHROM",
                               action='store_true', default=False, help="Flash rom.  Rom must be built previously.")
        parserObj.add_argument("--UPDATECONF", "--updateconf", "--UpdateConf",
                               dest="UPDATECONF", action='store_true', default=False,
                               help="Update Conf. Builders Conf files will be replaced with latest template files")
        parserObj.add_argument("--CLEAN", "--clean", "--CLEAN", dest="CLEAN",
                               action='store_true', default=False,
                               help="Clean. Remove all old build artifacts and intermediate files")
        parserObj.add_argument("--CLEANONLY", "--cleanonly", "--CleanOnly", dest="CLEANONLY",
                               action='store_true', default=False,
                               help="Clean Only. Do clean operation and don't build just exit.")
        parserObj.add_argument("--OUTPUTCONFIG", "--outputconfig", "--OutputConfig",
                               dest='OutputConfig', required=False, type=str,
                               help='Provide shell variables in a file')

    def RetrievePlatformCommandLineOptions(self, args):
        """Retrieve command line options from the argparser.

        Args:
            args (Namespace): namespace containing gathered args from argparser
        """
        self.OutputConfig = os.path.abspath(args.OutputConfig) if args.OutputConfig else None

        self.SkipBuild = args.SKIPBUILD
        self.SkipPreBuild = args.SKIPPREBUILD
        self.SkipPostBuild = args.SKIPPOSTBUILD
        self.Clean = args.CLEAN
        self.FlashImage = args.FLASHROM
        self.UpdateConf = args.UPDATECONF

        if (args.FLASHONLY):
            self.SkipPostBuild = True
            self.SkipBuild = True
            self.SkipPreBuild = True
            self.FlashImage = True
        elif (args.CLEANONLY):
            self.Clean = True
            self.SkipBuild = True
            self.SkipPreBuild = True
            self.SkipPostBuild = True
            self.FlashImage = False

    def Go(self, WorkSpace, PackagesPath, PInHelper, PInManager):
        """Core executable that performs all build steps."""
        self.env = shell_environment.GetBuildVars()
        self.mws = MultipleWorkspace()
        self.mws.setWs(WorkSpace, PackagesPath)
        self.edk2path = Edk2Path(WorkSpace, PackagesPath.split(os.pathsep))
        self.ws = WorkSpace
        self.pp = PackagesPath  # string using os.pathsep
        self.Helper = PInHelper
        self.pm = PInManager

        try:
            edk2_logging.log_progress("Start time: {0}".format(datetime.datetime.now()))
            start_time = time.perf_counter()

            self.Helper.DebugLogRegisteredFunctions()

            ret = self.SetEnv()
            if (ret != 0):
                logging.critical("SetEnv failed")
                return ret

            # clean
            if (self.Clean):
                edk2_logging.log_progress("Cleaning")
                ret = self.CleanTree()
                if (ret != 0):
                    logging.critical("Clean failed")
                    return ret

            # prebuild
            if (self.SkipPreBuild):
                edk2_logging.log_progress("Skipping Pre Build")
            else:
                ret = self.PreBuild()
                if (ret != 0):
                    logging.critical("Pre Build failed")
                    return ret

            # Output Build Environment to File - this is mostly for debug of build
            # issues or adding other build features using existing variables
            if (self.OutputConfig is not None):
                edk2_logging.log_progress("Writing Build Env Info out to File")
                logging.debug("Found an Output Build Env File: " + self.OutputConfig)
                self.env.PrintAll(self.OutputConfig)

            if (self.env.GetValue("GATEDBUILD") is not None) and (self.env.GetValue("GATEDBUILD").upper() == "TRUE"):
                ShouldGatedBuildRun = self.PlatformGatedBuildShouldHappen()
                logging.debug("Platform Gated Build Should Run returned: %s" % str(
                    ShouldGatedBuildRun))
                if (not self.SkipBuild):
                    self.SkipBuild = not ShouldGatedBuildRun
                if (not self.SkipPostBuild):
                    self.SkipPostBuild = not ShouldGatedBuildRun

            # build
            if (self.SkipBuild):
                edk2_logging.log_progress("Skipping Build")
            else:
                ret = self.Build()

                if (ret != 0):
                    logging.critical("Build failed")
                    return ret

            # postbuild
            if (self.SkipPostBuild):
                edk2_logging.log_progress("Skipping Post Build")
            else:
                ret = self.PostBuild()
                if (ret != 0):
                    logging.critical("Post Build failed")
                    return ret

            # flash
            if (self.FlashImage):
                edk2_logging.log_progress("Flashing Image")
                ret = self.FlashRomImage()
                if (ret != 0):
                    logging.critical("Flash Image failed")
                    return ret

        except Exception:
            logging.critical("Build Process Exception")
            logging.error(traceback.format_exc())
            return -1

        except SystemExit:
            logging.critical("Build Process Exit")
            logging.error(traceback.format_exc())
            return -1

        except KeyboardInterrupt:
            logging.critical("Build Cancelled by user")
            return -2

        finally:
            end_time = time.perf_counter()
            elapsed_time_s = int((end_time - start_time))
            edk2_logging.log_progress("End time: {0}\t Total time Elapsed: {1}".format(
                datetime.datetime.now(), datetime.timedelta(seconds=elapsed_time_s)))

        return 0

    def CleanTree(self, RemoveConfTemplateFilesToo=False):
        """Cleans the build directory.

        Args:
            RemoveConfTemplateFilesToo (bool): deletes conf files used for building makefiles
        """
        ret = 0
        # loop thru each build target set.
        edk2_logging.log_progress("Cleaning All Output for Build")

        d = self.env.GetValue("BUILD_OUTPUT_BASE")
        if (os.path.isdir(d)):
            logging.debug("Removing [%s]", d)
            # if the folder is opened in Explorer do not fail the entire Rebuild
            try:
                RemoveTree(d)
            except WindowsError as wex:
                logging.debug(wex)

        else:
            logging.debug("Directory [%s] already clean" % d)

        # delete the conf .dbcache
        # this needs to be removed in case build flags changed
        d = os.path.join(self.ws, "Conf", ".cache")
        if (os.path.isdir(d)):
            logging.debug("Removing [%s]" % d)
            RemoveTree(d)

        if (RemoveConfTemplateFilesToo):
            for a in ["target.txt", "build_rule.txt", "tools_def.txt"]:
                d = os.path.join(self.ws, "Conf", a)
                if (os.path.isfile(d)):
                    logging.debug("Removing [%s]" % d)
                    os.remove(d)

        return ret

    def Build(self):
        """Adds all arguments to the build command and runs it."""
        BuildType = self.env.GetValue("TARGET")
        edk2_logging.log_progress("Running Build %s" % BuildType)

        # set target, arch, toolchain, threads, and platform
        params = "-p " + self.env.GetValue("ACTIVE_PLATFORM")
        params += " -b " + BuildType
        params += " -t " + self.env.GetValue("TOOL_CHAIN_TAG")
        # Thread number is now optional and not set in default tianocore target.txt
        if self.env.GetValue("MAX_CONCURRENT_THREAD_NUMBER") is not None:
            params += " -n " + self.env.GetValue("MAX_CONCURRENT_THREAD_NUMBER")

        # Set the arch flags.  Multiple are split by space
        rt = self.env.GetValue("TARGET_ARCH").split(" ")
        for t in rt:
            params += " -a " + t

        # get the report options and setup the build command
        if (self.env.GetValue("BUILDREPORTING") == "TRUE"):
            params += " -y " + self.env.GetValue("BUILDREPORT_FILE")
            rt = self.env.GetValue("BUILDREPORT_TYPES").split(" ")
            for t in rt:
                params += " -Y " + t

        # add special processing to handle building a single module
        mod = self.env.GetValue("BUILDMODULE")
        if (mod is not None and len(mod.strip()) > 0):
            params += " -m " + mod
            edk2_logging.log_progress("Single Module Build: " + mod)
            self.SkipPostBuild = True
            self.FlashImage = False

        # attach the generic build vars
        buildvars = self.env.GetAllBuildKeyValues(BuildType)
        for key, value in buildvars.items():
            params += " -D " + key + "=" + value
        output_stream = edk2_logging.create_output_stream()

        env = shell_environment.ShellEnvironment()
        # WORKAROUND - Pin the PYTHONHASHSEED so that TianoCore build tools
        #               have consistent ordering. Addresses incremental builds.
        pre_build_env_chk = env.checkpoint()
        env.set_shell_var('PYTHONHASHSEED', '0')
        env.log_environment()

        edk2_build_cmd = self.env.GetValue("EDK_BUILD_CMD")
        if edk2_build_cmd is None:
            edk2_build_cmd = "build"
        logging.debug(f"The edk2 build command is {edk2_build_cmd}")

        edk2_build_params = self.env.GetValue("EDK_BUILD_PARAMS")
        if edk2_build_params is None:
            edk2_build_params = params
        logging.debug(f"Edk2 build parameters are {edk2_build_params}")

        ret = RunCmd(edk2_build_cmd, edk2_build_params, close_fds=False)
        # WORKAROUND - Undo the workaround.
        env.restore_checkpoint(pre_build_env_chk)

        problems = edk2_logging.scan_compiler_output(output_stream)
        edk2_logging.remove_output_stream(output_stream)
        for level, problem in problems:
            logging.log(level, problem)

        if (ret != 0):
            return ret

        return 0

    def PreBuild(self):
        """Performs internal PreBuild steps.

        This includes calling the platform overridable `PlatformPreBuild()`
        """
        edk2_logging.log_progress("Running Pre Build")
        #
        # Run the platform pre-build steps.
        #
        ret = self.PlatformPreBuild()

        if (ret != 0):
            logging.critical("PlatformPreBuild failed %d" % ret)
            return ret
        #
        # run all loaded UefiBuild Plugins
        #
        for Descriptor in self.pm.GetPluginsOfClass(IUefiBuildPlugin):
            rc = Descriptor.Obj.do_pre_build(self)
            if (rc != 0):
                if (rc is None):
                    logging.error(
                        "Plugin Failed: %s returned NoneType" % Descriptor.Name)
                    ret = -1
                else:
                    logging.error("Plugin Failed: %s returned %d" %
                                  (Descriptor.Name, rc))
                    ret = rc
                break  # fail on plugin error
            else:
                logging.debug("Plugin Success: %s" % Descriptor.Name)
        return ret

    def PostBuild(self):
        """Performs internal PostBuild steps.

        This includes calling the platform overridable `PlatformPostBuild()`.
        """
        edk2_logging.log_progress("Running Post Build")
        #
        # Run the platform post-build steps.
        #
        ret = self.PlatformPostBuild()

        if (ret != 0):
            logging.critical("PlatformPostBuild failed %d" % ret)
            return ret

        #
        # run all loaded UefiBuild Plugins
        #
        for Descriptor in self.pm.GetPluginsOfClass(IUefiBuildPlugin):
            rc = Descriptor.Obj.do_post_build(self)
            if (rc != 0):
                if (rc is None):
                    logging.error(
                        "Plugin Failed: %s returned NoneType" % Descriptor.Name)
                    ret = -1
                else:
                    logging.error("Plugin Failed: %s returned %d" %
                                  (Descriptor.Name, rc))
                    ret = rc
                break  # fail on plugin error
            else:
                logging.debug("Plugin Success: %s" % Descriptor.Name)

        return ret

    def SetEnv(self):
        """Performs internal SetEnv steps.

        This includes platform overridable `SetPlatformEnv()` and `SetPlatformEnvAfterTarget().
        """
        edk2_logging.log_progress("Setting up the Environment")
        shell_environment.GetEnvironment().set_shell_var("WORKSPACE", self.ws)
        shell_environment.GetBuildVars().SetValue("WORKSPACE", self.ws, "Set in SetEnv")

        if (self.pp is not None):
            shell_environment.GetEnvironment().set_shell_var("PACKAGES_PATH", self.pp)
            shell_environment.GetBuildVars().SetValue(
                "PACKAGES_PATH", self.pp, "Set in SetEnv")

        # process platform parameters defined in platform build file
        ret = self.SetPlatformEnv()
        if (ret != 0):
            logging.critical("Set Platform Env failed")
            return ret

        # set some basic defaults
        self.SetBasicDefaults()

        # Handle all the template files for workspace/conf/ Allow override
        TemplateDirList = [self.env.GetValue("EDK_TOOLS_PATH")]  # set to edk2 BaseTools
        PlatTemplatesForConf = self.env.GetValue("CONF_TEMPLATE_DIR")  # get platform defined additional path
        if (PlatTemplatesForConf is not None):
            PlatTemplatesForConf = self.edk2path.GetAbsolutePathOnThisSystemFromEdk2RelativePath(PlatTemplatesForConf)
            TemplateDirList.insert(0, PlatTemplatesForConf)
            logging.debug(f"Platform defined override for Template Conf Files: {PlatTemplatesForConf}")

        conf_dir = os.path.join(self.ws, "Conf")
        conf_mgmt.ConfMgmt().populate_conf_dir(conf_dir, self.UpdateConf, TemplateDirList)

        # parse target file
        ret = self.ParseTargetFile()
        if (ret != 0):
            logging.critical("ParseTargetFile failed")
            return ret

        # parse tools_def file
        ret = self.ParseToolsDefFile()
        if (ret != 0):
            logging.critical("ParseToolsDefFile failed")
            return ret

        # parse DSC file
        ret = self.ParseDscFile()
        if (ret != 0):
            logging.critical("ParseDscFile failed")
            return ret

        # parse FDF file
        ret = self.ParseFdfFile()
        if (ret != 0):
            logging.critical("ParseFdfFile failed")
            return ret

        # set build output base envs for all builds
        if self.env.GetValue("OUTPUT_DIRECTORY") is None:
            logging.warn("OUTPUT_DIRECTORY was not found, defaulting to Build")
            self.env.SetValue("OUTPUT_DIRECTORY", "Build", "default from uefi_build", True)

        # BUILD_OUT_TEMP is a path so the value should use native directory separators
        self.env.SetValue("BUILD_OUT_TEMP",
                          os.path.normpath(os.path.join(self.ws, self.env.GetValue("OUTPUT_DIRECTORY"))),
                          "Computed in SetEnv")

        target = self.env.GetValue("TARGET", None)
        if target is None:
            logging.error("Environment variable TARGET must be set to a build target.")
            logging.error("Review the 'CLI Env Guide' section provided when using stuart_build "\
                          "with the -help flag.")
            return -1

        self.env.SetValue("BUILD_OUTPUT_BASE", os.path.join(self.env.GetValue(
            "BUILD_OUT_TEMP"), target + "_" + self.env.GetValue("TOOL_CHAIN_TAG")), "Computed in SetEnv")

        # We have our build target now.  Give platform build one more chance for target specific settings.
        ret = self.SetPlatformEnvAfterTarget()
        if (ret != 0):
            logging.critical("SetPlatformEnvAfterTarget failed")
            return ret

        # set the build report file
        self.env.SetValue("BUILDREPORT_FILE", os.path.join(
            self.env.GetValue("BUILD_OUTPUT_BASE"), "BUILD_REPORT.TXT"), True)

        # set environment variables for the build process
        os.environ["EFI_SOURCE"] = self.ws

        # set critical platform Env Defaults if not set anywhere else in the build.
        for env_var in self.SetPlatformDefaultEnv():
            self.env.SetValue(env_var.name, env_var.default, "Default Critical Platform Env Value.")

        return 0

    def FlashRomImage(self):
        """Executes platform overridable `PlatformFlashImage()`."""
        return self.PlatformFlashImage()

    # -----------------------------------------------------------------------
    # Methods that will be overridden by child class
    # -----------------------------------------------------------------------

    @classmethod
    def PlatformPreBuild(self):
        """Perform Platform PreBuild Steps.

        Returns:
            (int): 0 on success, 1 on failure
        """
        return 0

    @classmethod
    def PlatformPostBuild(self):
        """Perform Platform PostBuild Steps.

        Returns:
            (int): 0 on success, 1 on failure
        """
        return 0

    @classmethod
    def SetPlatformEnv(self):
        """Set and read Platform Env variables.

        This is performed before platform files like the DSC and FDF have been parsed.

        !!! tip
            If a platform file (DSC, FDF, etc) relies on a variable set in
            the `UefiBuilder`, it must be set here, before the platform files
            have been parsed and values have been set.

        Returns:
            (int): 0 on success, 1 on failure
        """
        return 0

    @classmethod
    def SetPlatformEnvAfterTarget(self):
        """Set and read Platform Env variables after platform files have been parsed.

        Returns:
            (int): 0 on success, 1 on failure
        """
        return 0

    @classmethod
    def SetPlatformDefaultEnv(self) -> list[namedtuple]:
        """Sets platform default environment variables by returning them as a list.

        Variables returned from this method are printed to the command line when
        calling stuart_build with -h, --help. Variables added here should be
        reserved only those that are commonly overwritten in the command line for
        developers.

        Variables returned from this function are the last variables to be set,
        ensuring that default values are only added if none have been provided by
        other means.

        Returns:
            (list[namedtuple]): List of named tuples containing name, description, default
        """
        return []

    @classmethod
    def PlatformBuildRom(self):
        """Build the platform Rom.

        !!! tip
            Typically called by the platform in PlatformFlashImage. Not called
            automatically by the `UefiBuilder`.
        """
        return 0

    @classmethod
    def PlatformFlashImage(self):
        """Flashes the image to the system.

        Returns:
            (int): 0 on success, 1 on failure
        """
        return 0

    @classmethod
    def PlatformGatedBuildShouldHappen(self):
        """Specifies if a gated build should happen.

        Returns:
            (bool): True if gated build should happen, else False
        """
        return True

    # ------------------------------------------------------------------------
    #  HELPER FUNCTIONS
    # ------------------------------------------------------------------------
    #
    def ParseTargetFile(self):
        """Parses the target.txt file and adds values as env settings.

        "Sets them so they can be overriden.
        """
        conf_file_path = self.edk2path.GetAbsolutePathOnThisSystemFromEdk2RelativePath(
            "Conf", "target.txt")
        if os.path.isfile(conf_file_path):
            # parse TargetTxt File
            logging.debug("Parse Target.txt file")
            ttp = TargetTxtParser()
            ttp.ParseFile(conf_file_path)
            for key, value in ttp.Dict.items():
                # set env as overrideable
                self.env.SetValue(key, value, "From Target.txt", True)

            # Set the two additional edk2 common macros.  These will be resolved by now as
            # target.txt will set them if they aren't already set.
            self.env.SetValue("TOOLCHAIN", self.env.GetValue("TOOL_CHAIN_TAG"),
                              "DSC Spec macro - set based on tool_Chain_tag")
            # need to check how multiple arch are handled
            self.env.SetValue("ARCH", self.env.GetValue("TARGET_ARCH"), "DSC Spec macro - set based on target_arch")

        else:
            logging.error("Failed to find target.txt file")
            return -1

        return 0

    def ParseToolsDefFile(self):
        """Parses the tools_def.txt file and adds values as env settings.

        "Sets them so they can be overriden.
        """
        toolsdef_file_path = self.edk2path.GetAbsolutePathOnThisSystemFromEdk2RelativePath(
            "Conf", "tools_def.txt")
        if os.path.isfile(toolsdef_file_path):
            # parse ToolsdefTxt File
            logging.debug("Parse tools_def.txt file")
            tdp = TargetTxtParser()
            tdp.ParseFile(toolsdef_file_path)

            # Get the tool chain tag and then find the family
            # need to parse tools_def and find *_<TAG>_*_*_FAMILY
            # Example:  *_VS2019_*_*_FAMILY        = MSFT
            tool_chain = self.env.GetValue("TOOL_CHAIN_TAG", None)
            if tool_chain is None:
                logging.error("Environment variable TOOL_CHAIN_TAG must be set to a tool chain.")
                logging.error("Review the 'CLI Env Guide' section provided when using stuart_build "\
                              "with the -help flag.")
                return -1
            tag = "*_" + tool_chain + "_*_*_FAMILY"
            tool_chain_family = tdp.Dict.get(tag, "UNKNOWN")
            self.env.SetValue("FAMILY", tool_chain_family, "DSC Spec macro - from tools_def.txt")

        else:
            logging.error("Failed to find tools_def.txt file")
            return -1

        return 0

    def ParseDscFile(self):
        """Parses the active platform DSC file.

        This will get lots of variable info to be used in the build. This
        makes it so we don't have to define things twice.
        """
        if self.env.GetValue("ACTIVE_PLATFORM") is None:
            logging.error("The DSC file was not set. Please set ACTIVE_PLATFORM")
            return -1
        dsc_file_path = self.edk2path.GetAbsolutePathOnThisSystemFromEdk2RelativePath(
            self.env.GetValue("ACTIVE_PLATFORM"))
        if os.path.isfile(dsc_file_path):
            # parse DSC File
            logging.debug("Parse Active Platform DSC file: {0}".format(dsc_file_path))

            # Get the vars from the environment that are not build keys
            input_vars = self.env.GetAllNonBuildKeyValues()
            # Update with special environment set build keys
            input_vars.update(self.env.GetAllBuildKeyValues())
            dscp = DscParser().SetEdk2Path(self.edk2path).SetInputVars(input_vars)
            dscp.ParseFile(dsc_file_path)
            for key, value in dscp.LocalVars.items():
                # set env as overrideable
                self.env.SetValue(key, value, "From Platform DSC File", True)
        else:
            logging.error("Failed to find DSC file")
            return -1

        return 0

    def ParseFdfFile(self):
        """Parses the active platform FDF file.

        This will get lots of variable info to be used in the build. This makes
        it so we don't have to define things twice the FDF file usually comes
        from the Active Platform DSC file so it needs to be parsed first.
        """
        if (self.env.GetValue("FLASH_DEFINITION") is None):
            logging.debug("No flash definition set")
            return 0
        fdf_file_path = self.edk2path.GetAbsolutePathOnThisSystemFromEdk2RelativePath(
            self.env.GetValue("FLASH_DEFINITION"))
        if os.path.isfile(fdf_file_path):
            # parse the FDF file- fdf files have similar syntax to DSC and therefore parser works for both.
            logging.debug("Parse Active Flash Definition (FDF) file")

            # Get the vars from the environment that are not build keys
            input_vars = self.env.GetAllNonBuildKeyValues()
            # Update with special environment set build keys
            input_vars.update(self.env.GetAllBuildKeyValues())
            fdf_parser = FdfParser().SetEdk2Path(self.edk2path).SetInputVars(input_vars)
            fdf_parser.ParseFile(fdf_file_path)
            for key, value in fdf_parser.LocalVars.items():
                self.env.SetValue(key, value, "From Platform FDF File", True)

        else:
            logging.error("Failed to find FDF file")
            return -2

        return 0

    def SetBasicDefaults(self):
        """Sets default values for numerous build control flow variables."""
        self.env.SetValue("WORKSPACE", self.ws, "DEFAULT")
        if (self.pp is not None):
            self.env.SetValue("PACKAGES_PATH", self.pp, "DEFAULT")
        return 0



# @file var_dict.py
# This module contains code for a special overridable dictionary.
# This stores most of the build configuration data and allows
# extensive config sharing for the build process, pre-build, and
# post-build.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A special overridable dictionary.

Stores most of the build configuration data and allows extensive config
sharing for the build process, pre-build, and post-build.
"""

import logging
from random import choice
from string import ascii_letters


class EnvEntry(object):
    """A single Environment Variable entry for VarDict.

    Attributes:
        Value (obj): The value to store in the dictionary
        Comment (str): A debug comment specifying where / how the value was set
        overridable (bool): If the value can be overwritten in the future
    """

    def __init__(self, value, comment, overridable=False):
        """Inits an entry with the specified values."""
        self.Value = value
        self.Comment = comment
        self.Overrideable = overridable

    def PrintEntry(self, f=None):
        """Prints the value.

        Args:
            f (str): a file to write to instead of the terminal.
        """
        print("Value: %s" % self.Value, file=f)
        print("Comment: %s" % self.Comment, file=f)
        if (self.Overrideable):
            print("Value overridable", file=f)
        print("**********************", file=f)
    #
    # Function used to override the value if option allows it
    #

    def SetValue(self, value, comment, overridable=False):
        """Sets the value of the entry if it os overridable.

        Args:
            value (obj): value to set
            comment (str): A debug comment specifying where / how the value was set
            overridable (bool): If the value can be overwritten in the future

        !!! warning
            Even if you set a value as overridable=False, another entity can
            call `AllowOverride()` and change the value anyway.
        """
        if (value == self.Value) and (overridable == self.Overrideable):
            return True

        if (not self.Overrideable):
            logging.debug("Can't set value [%s] as it isn't overrideable. Previous comment %s" % (
                value, self.Comment))
            return False

        self.Value = value
        self.Comment = comment
        self.Overrideable = overridable
        return True

    def AllowOverride(self):
        """Allows the value to be overwritten in the future."""
        self.Overrideable = True
        return True

    def GetValue(self):
        """Returns the value."""
        return self.Value


class VarDict(object):
    """An overridable dictionary to store build configuration data."""

    def __init__(self):
        """Inits an empty VarDict."""
        self.Logger = logging.getLogger("EnvDict")
        self.Dstore = {}  # a set of envs

    def GetEntry(self, key):
        """Returns an entry in the Dstore Dict."""
        return self.Dstore.get(key.upper())

    def __copy__(self):
        """Copies data into a new VarDict."""
        new_copy = VarDict()
        new_copy.Logger = self.Logger

        new_copy.Dstore = {}
        for key in self.Dstore:
            entry = self.GetEntry(key)
            value = entry.Value
            comment = entry.Comment
            override = entry.Overrideable
            new_copy.SetValue(key, value, comment, override)
        return new_copy

    def GetValue(self, k, default=None):
        """Gets a value from the variable dictionary that was set during build.

        !!! note
            Values set in DSC, FDF, and CLI stored as strings

        Args:
            k (str): The key the value was stored as
            default (varied): default value if key is not present

        Returns:
            (varied): The value of the key, if present, else default value
        """
        if (k is None):
            logging.debug(
                "GetValue - Invalid Parameter key is None.")
            return None

        key = k.upper()
        en = self.GetEntry(key)
        if (en is not None):
            self.Logger.debug("Key %s found.  Value %s" % (key, en.GetValue()))
            return en.GetValue()
        else:
            self.Logger.debug("Key %s not found" % key)
            return default

    def SetValue(self, k, v, comment, overridable=False):
        """Sets an environment variable to be used throughout the build.

        Args:
            k (str): The key to store the value under
            v (varied | None): The value to store as a string, or None to store
                a non valued build variable
            comment (str): A comment to show where / how the variable was stored.
                Useful for debugging
            overridable (bool): Specifies if the variable is allowed to be override
                elsewhere in the build

        Returns:
            (bool): If the variable was successfully stored or not
        """
        key = k.upper()
        en = self.GetEntry(key)
        if v is None:
            value = ''.join(choice(ascii_letters) for _ in range(20))
        else:
            value = str(v)
        self.Logger.debug("Trying to set key %s to value %s" % (k, v))
        if (en is None):
            # new entry
            en = EnvEntry(value, comment, overridable)
            self.Dstore[key] = en
            return True

        return en.SetValue(value, comment, overridable)

    def AllowOverride(self, k):
        """Forces the key/value pair to be overridable.

        Note: Even if overridable was specifically set to False,
        it still allows it.

        Args:
            k (str): The key the value was stored as

        Returns:
            (bool): if the key existed or not
        """
        key = k.upper()
        en = self.GetEntry(key)
        if (en is not None):
            self.Logger.warning("Allowing Override for key %s" % k)
            en.AllowOverride()
            return True
        return False

    def GetBuildValue(self, key, BuildType=None):
        """Get a build var value for given key and buildtype.

        !!! tip
            Build vars are defined by vars that start with BLD_

            BLD_*_<YOUR VAR HERE> means all build types

            BLD_DEBUG_<YOUR VAR HERE> means build of debug type

            BLD_RELEASE_<YOUR VAR HERE> means build of release type

            etc

        Args:
            key (str): The key the value was stored as
            BuildType (str): DEBUG/RELEASE

        Returns:
            (str): The value of the key, if present, else None
        """
        rv = None

        if (BuildType is None):
            BuildType = self.GetValue("TARGET")

        if (BuildType is None):
            logging.debug(
                "GetBuildValue - Invalid Parameter BuildType is None and Target Not set. Key is: " + key)
            return None

        if (key is None):
            logging.debug(
                "GetBuildValue - Invalid Parameter key is None. BuildType is: " + BuildType)
            return None

        ty = BuildType.upper().strip()
        tk = key.upper().strip()
        # see if specific
        k = "BLD_" + ty + "_" + tk
        rv = self.GetValue(k)
        if (rv is None):
            # didn't fine build type specific so check for generic
            k = "BLD_*_" + tk
            rv = self.GetValue(k)

        # return value...if not found should return None
        return rv

    def GetAllBuildKeyValues(self, BuildType=None):
        """Gets a dictionary for all build vars.

        !!! tip
            Build vars are defined by vars that start with BLD_

            BLD_*_<YOUR VAR HERE> means all build types

            BLD_DEBUG_<YOUR VAR HERE> means build of debug type

            BLD_RELEASE_<YOUR VAR HERE> means build of release type

            etc

        Args:
            BuildType (:obj:`str`, optional): DEBUG/RELEASE

        Returns:
            (dict): all keys, values in the environment which are build keys

        """
        returndict = {}
        if (BuildType is None):
            BuildType = self.GetValue("TARGET")

        if (BuildType is None):
            logging.debug(
                "GetAllBuildKeyValues - Invalid Parameter BuildType is None and Target Not Set.")
            return returndict

        ty = BuildType.upper().strip()
        logging.debug("Getting all build keys for build type " + ty)

        # get all the generic build options
        for key, value in self.Dstore.items():
            if (key.startswith("BLD_*_")):
                k = key[6:]
                returndict[k] = value.GetValue()

        # will override with specific for this build type
        # figure out offset part of key name to strip
        ks = len(ty) + 5
        for key, value in self.Dstore.items():
            if (key.startswith("BLD_" + ty + "_")):
                k = key[ks:]
                returndict[k] = value.GetValue()

        return returndict

    def GetAllNonBuildKeyValues(self):
        """Returns a dict of non Build Key values.

        Return a copy of the dictionary of all keys, values in the environment
        which are not Build Keys.
        """
        returndict = {}
        # get all the generic build options
        for key, value in self.Dstore.items():
            if (not key.startswith("BLD_")):
                returndict[key] = value.GetValue()
        return returndict

    def PrintAll(self, fp=None):
        """Prints all variables.

        If fp is not none, writes to a fp also

        Args:
            fp (str): file pointer to print to
        """
        f = None
        if (fp is not None):
            f = open(fp, 'a+')
        for key, value in self.Dstore.items():
            print("Key = %s" % key, file=f)
            value.PrintEntry(f)
        if (f):
            f.close()



# @file capsule_tool.py
# This module contains the CLI interface for easily creating EDK FMP UEFI capsules
# from payload files.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""CLI interface for creating EDK FMP UEFI Capsules from payload files."""

import argparse
import copy
import logging
import os
import sys

import yaml

from edk2toolext.capsule import capsule_helper, signing_helper

TOOL_DESCRIPTION = """
EDK Capsule Tool is a command-line tool to assist with the production of UEFI
capsules. It takes in a payload and any number of options to produce a
properly constructed capsule that can be parsed by the EDK2 FMP Capsule
infrastructure.

An example call might look like:
%s -o /path/to/config.yaml -ds key_file="/other/signer/key/file.pfx"
    /path/to/payload.bin /path/to/output
""" % (os.path.basename(sys.argv[0]),)


def get_cli_options(args=None):
    """Parse the primary options from the command line."""
    parser = argparse.ArgumentParser(description=TOOL_DESCRIPTION, formatter_class=argparse.RawDescriptionHelpFormatter)

    # Add the group for the signer specifier.
    # NOTE: At least one signer type is required!
    signer_group = parser.add_mutually_exclusive_group(required=True)
    signer_group.add_argument(
        '--builtin_signer', choices=[signing_helper.PYOPENSSL_SIGNER, signing_helper.SIGNTOOL_SIGNER])
    signer_group.add_argument(
        '--local_signer', help='a filesystem path to a python module that can be loaded as the active signer')
    signer_group.add_argument(
        '--module_signer', help='a python dot-path to a signer module that can be loaded from the current pypath')

    options_help = 'add an option to the corresponding set. format is <option_name>=<option_value>'
    parser.add_argument('-dc', action='append', dest='capsule_options', type=str, default=[], help=options_help)
    parser.add_argument('-ds', action='append', dest='signer_options', type=str, default=[], help=options_help)

    parser.add_argument('-o', dest='options_file', type=argparse.FileType('r'),
                        help='a filesystem path to a json/yaml file to load with default options. will be overriden by any options parameters') # noqa
    parser.add_argument('-f', dest='save_final_options', default=False, action='store_true',
                        help='optional flag to request that final tool options be saved in a file in the output directory')                     # noqa

    parser.add_argument('capsule_payload', type=argparse.FileType('rb'),
                        help='a filesystem path to the binary payload for the capsule')
    parser.add_argument('output_dir',
                        help='a filesystem path to the directory to save output files. if directory does not exist, entire directory path will be created. if directory does exist, contents will be updated based on the capsule_options') # noqa

    return parser.parse_args(args=args)


def load_options_file(in_file):
    """Loads a yaml file into a dictionary and returns it."""
    if not hasattr(in_file, 'read'):
        return None

    return yaml.safe_load(in_file)


def update_options(file_options, capsule_options, signer_options):
    """Takes in a pre-loaded options dictionary and adds all corresponding options from the command line.

    Command line options will be organized by type (into capsule_options or
    signer_options) and in lists of strings that look like
    '<option_name>=<option_value>'.
    """
    if file_options is not None:
        updated_options = copy.copy(file_options)
    else:
        updated_options = {'capsule': {}, 'signer': {}}

    # Update all the capsule options.
    for option in capsule_options:
        (key, value) = option.split('=')
        updated_options['capsule'][key] = value

    # Update all the signer options.
    for option in signer_options:
        (key, value) = option.split('=')
        updated_options['signer'][key] = value

    return updated_options


def main():
    """Main entry point into the Capsule Tool."""
    args = get_cli_options()
    final_options = update_options(load_options_file(args.options_file), args.capsule_options, args.signer_options)

    # Verify minimum capsule options.
    required_capsule_options = ('fw_name', 'fw_version', 'lsv_version', 'fw_version_string',
                                'provider_name', 'fw_description', 'esrt_guid')
    missing_capsule_options = tuple(option for option in required_capsule_options
                                    if option not in final_options['capsule'])
    if len(missing_capsule_options) > 0:
        logging.error("Missing required capsule options: " + ", ".join(missing_capsule_options) + "!")
        logging.error("Options MUST be provided in either the options file or on the command line.")
        sys.exit(1)

    # Next, we're gonna need a signer.
    if args.builtin_signer is not None:
        signer = signing_helper.get_signer(args.builtin_signer)
    elif args.module_signer is not None:
        signer = signing_helper.get_signer(signing_helper.PYPATH_MODULE_SIGNER, args.module_signer)
    elif args.local_signer is not None:
        signer = signing_helper.get_signer(signing_helper.LOCAL_MODULE_SIGNER, args.local_signer)

    # Now, build the capsule.
    uefi_capsule_header = capsule_helper.build_capsule(
        args.capsule_payload.read(),
        final_options['capsule'],
        signer,
        final_options['signer']
    )

    # Save the capsule.
    capsule_helper.save_capsule(uefi_capsule_header, final_options['capsule'], args.output_dir)

    # Build the INF file.
    capsule_helper.create_inf_file(final_options['capsule'], args.output_dir)

    # Build the CAT file.
    capsule_helper.create_cat_file(final_options['capsule'], args.output_dir)

    # If requested, save the final options for provenance.
    if args.save_final_options:
        final_options_file = os.path.join(args.output_dir, 'Final_Capsule_Options.yaml')
        with open(final_options_file, 'w') as options_file:
            yaml.dump(final_options, options_file, indent=2)


if __name__ == '__main__':
    main()



# @file version_aggregator facilitates the collection of information
# regarding the tools, binaries, submodule configuration used in a build
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Used to facilitate the collection of information.

Used to facilitate the collection of information regarding the tools,
binaries, submodule configuration used in a build.
"""
import copy
import logging
from enum import Enum

VERSION_AGGREGATOR = None


class version_aggregator(object):
    """Used to facilitate the collection of information.

    Used to facilitate the collection of information regarding the tools,
    binaries, submodule configuration used in a build.
    """
    def __init__(self):
        """Inits an empty verion aggregator."""
        super(version_aggregator, self).__init__()
        self._Versions = {}
        self._logger = logging.getLogger("version_aggregator")

    def ReportVersion(self, key, value, versionType, path=None):
        """Report the version of something.

        Args:
            key (str): the name of what you are reporting.
            value (str): The value of what you are reporting.
            versionType (str): The method of categorizing what is being reported. See VersionTypes for details.
            path (str): the associated path.
        """
        if key in self._Versions:
            old_version = self._Versions[key]
            if old_version["version"] == value and old_version["path"] == path:
                self._logger.info(f"version_aggregator: {key} re-registered at {path}")
            else:
                error = "version_aggregator: {0} key registered with a different value\n\t" \
                        "Old:{1}@{3}\n\tNew:{2}@{4}\n".format(
                            key, old_version["version"], value, old_version["path"], path)
                self._logger.error(error)
                raise ValueError(error)
            return

        self._Versions[key] = {
            "name": key,
            "version": value,
            "type": versionType.name,
            "path": path
        }
        self._logger.debug("version_aggregator logging version: {0}".format(str(self._Versions[key])))

    def Print(self):
        """Prints out the current information from the version aggregator."""
        for version_key in self._Versions:
            version = self._Versions[version_key]
            print(f"{version['type']} - {version['name']}: {version['version']}")
        if len(self._Versions) == 0:
            print("VERSION AGGREGATOR IS EMPTY")

    def GetAggregatedVersionInformation(self):
        """Returns a copy of the aggregated information."""
        return copy.deepcopy(self._Versions)

    def Reset(self):
        """Resets all versions."""
        self._Versions = {}


class VersionTypes(Enum):
    """Enumerator representing the different version types for recording.

    Attributes:
        COMMIT: the commit hash of a repository.
        BINARY: pre-packaged binary that is distributed with a version number.
        TOOL: the version number of a tool that was used during the build process.
        INFO: miscellaneous information.
        PIP: a python pip package.
    """
    TOOL = 1
    COMMIT = 2
    BINARY = 3
    INFO = 4
    PIP = 5


def GetVersionAggregator():
    """Returns a singleton instance of this class for global use."""
    global VERSION_AGGREGATOR

    if VERSION_AGGREGATOR is None:
        logging.debug("Setting up version aggregator")
        VERSION_AGGREGATOR = version_aggregator()

    return VERSION_AGGREGATOR


def ResetVersionAggregator():
    """Resets the version Aggregator singleton."""
    GetVersionAggregator().Reset()



# @file uefi_helper_plugin
# Plugin that supports adding Extension or helper methods
# to the build environment
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Plugin that supports adding Extension or helper methods to the build environment."""
import logging


class IUefiHelperPlugin(object):
    """The class that should be subclassed when creating a UEFI Helper Plugin."""

    def RegisterHelpers(self, obj):
        """Allows a plugin to register its functions.

        !!! tip
            ```obj.Register()```

        Args:
            obj (HelperFunctions): HelperFunctions object that allows functional registration
        """


class HelperFunctions(object):
    """A class that contains all registed functions.

    Attributes:
        RegisteredFunctions(dict): registered functions
    """
    def __init__(self):
        """Initializes instance."""
        self.RegisteredFunctions = {}

    def DebugLogRegisteredFunctions(self):
        """Logs all registered functions and their source path.

        Uses logging.debug to write all registered functions and their source path.
        """
        logging.debug("Logging all Registered Helper Functions:")
        for name, file in self.RegisteredFunctions.items():
            logging.debug("  Function %s registered from file %s", name, file)
        logging.debug("Finished logging %d functions",
                      len(self.RegisteredFunctions))

    def Register(self, name, function, filepath):
        """Registers a plugin.

        Plugins that want to register a helper function should call
        this routine for each function

        Args:
            name (str): name of the function
            function (func): function being registered
            filepath (str): filepath to this file

        !!! tip
            ```os.path.abspath(__file__)```
        """
        if (name in self.RegisteredFunctions.keys()):
            raise Exception("Function %s already registered from plugin file %s.  Can't register again from %s" % (
                name, self.RegisteredFunctions[name], filepath))
        setattr(self, name, function)
        self.RegisteredFunctions[name] = filepath

    def HasFunction(self, name):
        """Returns if a function exists.

        Args:
            name (str): name of the function

        Returns:
            (bool): if the function is registered or not.
        """
        if (name in self.RegisteredFunctions.keys()):
            return True
        else:
            return False

    def LoadFromPluginManager(self, pm):
        """Load all IUefiHelperPlugins into the class.

        Uses the PluginManager class to get all IUefiHelperPlugins in the environment,
        then stores them all in a dict.

        Args:
            pm (PluginManager): class holding all plugins

        Returns:
            (int): number of plugins that failed to be loaded.
        """
        error = 0
        for Descriptor in pm.GetPluginsOfClass(IUefiHelperPlugin):
            logging.info(Descriptor)
            logging.debug("Helper Plugin Register: %s", Descriptor.Name)
            try:
                Descriptor.Obj.RegisterHelpers(self)
            except Exception as e:
                logging.warning(
                    "Unable to register {0}".format(Descriptor.Name))
                logging.error(e)
                error += 1
        return error



##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
# spell-checker:ignore rtags
"""Omnicache tool for lessening network usage and time of cloning repositories."""
import argparse
import datetime
import logging
import os
import sys
import uuid
from io import StringIO

import yaml
from edk2toollib.utility_functions import RunCmd

from edk2toolext import edk2_logging

# Omnicache version 0.11+ design notes:
# 1. Only a single remote per URL will be permitted.
# 2. Remote names inside omnicache will be UUIDs generated when a new URL is added.
# 3. Users may specify names for remotes, but they will be treated simply as "display names" without git significance.
# 4. As many objects as possible will be cached for any given remote; so all remotes will fetch tags.
# 4a. To avoid tag name collisions between remotes, tags will be fetched into per-remote namespaces.
# 5. Older omnicaches will be updated to meet the above design points when first encountered by the script.
# 5a. This implementation takes care that the underlying git construction is compatible (but not performant) when an
#     older omnicache interacts with it. Older omnicache should not crash when encountering a newer omnicache directory
#     but may take a lot of slow and unecessary actions to "re-initialize" it.

OMNICACHE_VERSION = "0.12"

# Prior to version 0.11, an "omnicache.yaml" file in the root of the omnicache repo was used to manage the cache
# configuration. Starting with 0.11, git config entries are used directly. If the file below is present, it will be
# removed to force re-initialization of the omnicache by older omnicache instances. (e.g. if an old branch with a prior
# implementation of omnicache is being built and uses the same cache directory)
PRE_0_11_OMNICACHE_FILENAME = "omnicache.yaml"


class Omnicache():
    """Class for managing an omnicache instance."""
    def __init__(self, cachepath, create=False, convert=True):
        """Initializes an omnicache.

        Args:
            cachepath (str): path to the omnicache git repo
            create (bool):  create a new omnicache instance if it doesn't exist
            convert (bool): if an old (but compatible) omnicache already exists, at cachepath, convert it to this
                version of Omnicache
        """
        self.path = cachepath
        self._InvalidateUrlLookupCache()

        (valid, isConversionCandidate) = self._ValidateOmnicache()
        if (not valid):
            if create and not isConversionCandidate:
                self._InitOmnicache()
            elif (convert and isConversionCandidate):
                self._ConvertOmnicache()
            else:
                logging.critical("Omnicache at {0} not valid, and cannot create/convert.".format(cachepath))
                raise RuntimeError("Invalid cache path: {0}".format(cachepath))

    def _ValidateOmnicache(self):
        """Validates whether self.path has a valid omnicache instance.

        Returns:
            (bool, bool): valid, convertible

        NOTE: "valid" is True if a compatible Omnicache exists at self.path
        NOTE: "convertible" is True if an older Omnicache that can be converted exists at self.path.
        """
        logging.info("Checking if {0} is valid omnicache.".format(self.path))
        if (not os.path.isdir(self.path)):
            logging.debug("{0} does not exist - not valid (not convertible).".format(self.path))
            return (False, False)
        out = StringIO()
        ret = RunCmd("git", "rev-parse --is-bare-repository", workingdir=self.path, outstream=out)
        if (ret != 0):
            logging.debug("{0} error getting repo state - not valid (not convertible).".format(self.path))
            return (False, False)
        if (out.getvalue().strip().lower() == "true"):
            if (os.path.exists(os.path.join(self.path, PRE_0_11_OMNICACHE_FILENAME))):
                logging.debug("{0} - old config file present. not valid (is convertible).".format(self.path))
                return (False, True)
            out = StringIO()
            ret = RunCmd("git", "config --local omnicache.metadata.version", workingdir=self.path, outstream=out)
            if (ret != 0):
                logging.debug("{0} - error retrieving omnicache version. not valid (is convertible).".format(self.path))
                return (False, True)
            if (out.getvalue().strip() == OMNICACHE_VERSION):
                logging.debug("{0} - matching omnicache version. valid (convertible don't care).".format(self.path))
                return (True, True)
            else:
                logging.debug("{0} - non-matching omnicache version. not valid (is convertible).".format(self.path))
                return (False, True)
        logging.debug("{0} - not a bare repo. not valid (not convertible).".format(self.path))
        return (False, False)

    def _InitOmnicache(self):
        """Initializes a new omnicache instance."""
        logging.critical("Initializing Omnicache in {0}".format(self.path))
        os.makedirs(self.path, exist_ok=True)
        ret = RunCmd("git", "init --bare", workingdir=self.path)
        if (ret != 0):
            return ret
        # by default, git fetch is single-threaded. This configuration allows git to use a "reasonable" default
        # (presently equal to the number of cpus) to execute the fetch in parallel.
        ret = RunCmd("git", "config --local fetch.parallel 0", workingdir=self.path)
        if (ret != 0):
            return ret

        return RunCmd("git",
                      "config --local omnicache.metadata.version {0}".format(OMNICACHE_VERSION),
                      workingdir=self.path)

    def _ConvertOmnicache(self):
        """Converts an existing bare git repo from a previous omnicache version to the current omnicache version."""
        logging.info("Converting Omnicache in {0} to latest format.".format(self.path))
        if (os.path.exists(os.path.join(self.path, PRE_0_11_OMNICACHE_FILENAME))):
            os.remove(os.path.join(self.path, PRE_0_11_OMNICACHE_FILENAME))
        remotes = Omnicache.GetRemotes(self.path)
        logging.info("Renaming non-UUID remotes with UUID.")
        for (name, _) in remotes.items():
            if (not Omnicache._IsValidUuid(name)):
                logging.info("Converting remote {0} to UUID".format(name))
                # Rename the remote with a valid UUID
                newName = str(uuid.uuid4())
                ret = RunCmd("git", "remote rename {0} {1}".format(name, newName), workingdir=self.path)
                if (ret != 0):
                    # rename failed; try removal.
                    logging.warn("Failed to rename {0}. Attempting to remove it.".format(name))
                    ret = RunCmd("git", "remote remove {0}".format(name), workingdir=self.path)
                    if (ret != 0):
                        logging.critical("Failed to rename or remove {0} - skipping.".format(name))
                        continue
                # Remove previous fetch config entries and regenerate them. Proceed to create new ones even if it fails.
                RunCmd("git", "config --local --unset-all remote.{0}.fetch".format(newName), workingdir=self.path)
                RunCmd("git",
                       "config --local --add remote.{0}.fetch +refs/heads/*:refs/remotes/{0}/*".format(newName),
                       workingdir=self.path)
                RunCmd("git",
                       "config --local --add remote.{0}.fetch refs/tags/*:refs/rtags/{0}/*".format(newName),
                       workingdir=self.path)
                # Add the original name as a display name
                RunCmd("git",
                       "config --local omnicache.{0}.displayname {1}".format(newName, name),
                       workingdir=self.path)
                logging.info("Remote {0} converted to {1}.".format(name, newName))
        # delete any tags in the global name space (older omnicaches fetched all tags into the global tag namespace)
        logging.info("Removing global tags")
        out = StringIO()
        RunCmd("git", "tag -l", workingdir=self.path, outstream=out)
        tags = " ".join(out.getvalue().splitlines())
        RunCmd("git", "tag -d {0}".format(tags), workingdir=self.path)
        # remove duplicates
        logging.info("Removing remotes with duplicate URLs")
        knownUrls = []
        remotes = Omnicache.GetRemotes(self.path)
        for (name, url) in remotes.items():
            if (url not in knownUrls):
                logging.info("Retaining remote {0} with unique URL {1}".format(name, url))
                knownUrls.append(url)
            else:
                logging.info("Removing remote {0} with duplicate URL {1}".format(name, url))
                RunCmd("git", "remote remove {0}".format(name), workingdir=self.path)
                RunCmd("git", "config --local --unset omnicache.{0}.displayname".format(name), workingdir=self.path)
        # by default, git fetch is single-threaded. This configuration allows git to use a "reasonable" default
        # (presently equal to the number of cpus) to execute the fetch in parallel.
        ret = RunCmd("git", "config --local fetch.parallel 0", workingdir=self.path)
        if (ret != 0):
            return ret
        # write current omnicache version into cache
        logging.info("Writing Omnicache version")
        return RunCmd("git",
                      "config --local omnicache.metadata.version {0}".format(OMNICACHE_VERSION),
                      workingdir=self.path)

    def _RefreshUrlLookupCache(self):
        """Refreshes the URL lookup cache."""
        if (len(self.urlLookupCache) == 0):
            logging.info("Regenerating URL lookup cache.")
            out = StringIO()
            ret = RunCmd("git", r"config --local --get-regexp remote\..*?\.url", workingdir=self.path, outstream=out)
            if (ret != 0):
                return None
            # output is in the form: remote.<name>.url <url>
            for remote in out.getvalue().splitlines():
                self.urlLookupCache[remote.split()[1]] = ".".join(remote.split()[0].split(".")[1:-1])

    def _InvalidateUrlLookupCache(self):
        """Invalidates the URL lookup cache."""
        logging.debug("Invalidating URL lookup cache.")
        self.urlLookupCache = {}

    def _LookupRemoteForUrl(self, url):
        """Returns the git remote name for the specified URL, or None if it doesn't exist."""
        self._RefreshUrlLookupCache()
        if (url in self.urlLookupCache):
            return self.urlLookupCache[url]
        return None

    def AddRemote(self, url, name=None):
        """Adds a remote for the specified URL to the omnicache.

        Args:
            url (str): URL for which to add a remote
            name(str, optional): provides a "display name" to be associated with this remote.
        """
        # if we already have this remote (i.e. a remote with this URL exists), then just update.
        if (self._LookupRemoteForUrl(url) is not None):
            return self.UpdateRemote(url, newName=name)
        # otherwise create it.
        logging.info("Adding new remote for url {0}".format(url))
        newName = str(uuid.uuid4())
        ret = RunCmd("git", "remote add {0} {1}".format(newName, url), workingdir=self.path)
        if (ret != 0):
            return ret
        self._InvalidateUrlLookupCache()
        # add display name, if specified
        if (name is not None):
            ret = RunCmd("git",
                         "config --local omnicache.{0}.displayname {1}".format(newName, name),
                         workingdir=self.path)
            if (ret != 0):
                return ret
        # add a special fetch refspec to fetch remote tags into a per-remote local namespace.
        return RunCmd("git",
                      "config --local --add remote.{0}.fetch refs/tags/*:refs/rtags/{0}/*".format(newName),
                      workingdir=self.path)

    def RemoveRemote(self, url):
        """Removes the remote for the specified url from the cache."""
        name = self._LookupRemoteForUrl(url)
        if (name is None):
            logging.critical("Failed to remove node for url {0}: such a remote does not exist.".format(url))
            return 1
        logging.info("Removing remote for url {0}".format(url))
        ret = RunCmd("git", "remote remove {0}".format(name), workingdir=self.path)
        self._InvalidateUrlLookupCache()
        return ret

    def UpdateRemote(self, oldUrl, newUrl=None, newName=None):
        """Updates the remote.

        Args:
            oldUrl (str): current url for the remote
            newUrl (str, optional): updates the remote to point to the new URL.
            newName (str, optional): updates the "displayname" for the remote.
        """
        remote = self._LookupRemoteForUrl(oldUrl)
        if (remote is None):
            logging.critical("Failed to update node for url {0}: such a remote does not exist.".format(oldUrl))
            return 1
        if (newName is not None):
            logging.info("Updating display name for url {0} to {1}".format(oldUrl, newName))
            ret = RunCmd("git",
                         "config --local omnicache.{0}.displayname {1}".format(remote, newName),
                         workingdir=self.path)
            if (ret != 0):
                return ret
        if (newUrl is not None):
            logging.info("Updating url {0} to {1}".format(oldUrl, newUrl))
            ret = RunCmd("git", "remote set-url {0} {1}".format(remote, newUrl), workingdir=self.path)
            self._InvalidateUrlLookupCache()
            if (ret != 0):
                return ret

        return 0

    def Fetch(self, jobs=0):
        """Fetches all remotes."""
        logging.info("Fetching all remotes.")
        self._RefreshUrlLookupCache()
        # Tricky: we pass no-tags here, since we set up custom fetch refs for tags on a per-remote basis. This prevents
        # git from fetching the first set of tags into the global namespace.
        if (jobs != 0):
            return RunCmd("git", "fetch --all -j {0} --no-tags".format(jobs), workingdir=self.path)
        else:
            return RunCmd("git", "fetch --all --no-tags", workingdir=self.path)

    def GetRemoteData(self):
        """Gets Remote Data.

        Returns:
            (dict(dict)): {"<uuid>":{"url":<url>, "displayname":<displayname>}}

        Note: "displayname" may not be present if the remote has no display name.
        """
        logging.info("Retrieving all remote data")
        self._RefreshUrlLookupCache()
        remoteData = {}
        for url in self.urlLookupCache.keys():
            remoteData[self.urlLookupCache[url]] = {"url": url}

        out = StringIO()
        ret = RunCmd("git",
                     r"config --local --get-regexp omnicache\..*?\.displayname",
                     workingdir=self.path,
                     outstream=out)
        if (ret != 0):
            return remoteData

        for displayName in out.getvalue().splitlines():
            remoteName = displayName.split()[0].split(".")[1]
            if (remoteName in remoteData.keys()):
                remoteData[remoteName].update({"displayname": displayName.split()[1]})
        return remoteData

    def List(self):
        """Prints the current set of remotes."""
        print("List OMNICACHE content:\n")
        remoteData = self.GetRemoteData()
        if (len(remoteData) == 0):
            print("No remotes.")
        for (name, data) in remoteData.items():
            print("Id {0}: {1}".format(name, str(data)))

    @staticmethod
    def GetRemotes(path):
        """Gets the remotes for the git repository at the specified path.

        Returns:
            (dict): {<remote>:<url>}
        """
        logging.info("Retreiving remotes")
        remotes = {}
        out = StringIO()

        ret = RunCmd("git", "remote -v", workingdir=path, outstream=out)
        if (ret != 0):
            return remotes

        # Note: this loop assumes fetch and push URLs will be identical. If not, the last URL output will be the result.
        for remote in out.getvalue().splitlines():
            remoteInfo = remote.split()
            remotes[remoteInfo[0]] = remoteInfo[1]

        return remotes

    @staticmethod
    def _IsValidUuid(val):
        """Returns whether the input is valid UUID."""
        try:
            uuid.UUID(str(val))
            return True
        except ValueError:
            pass
        return False


def ProcessInputConfig(omnicache, input_config):
    """Adds a list of remotes from a YAML config file to the omnicache."""
    logging.info("Adding remotes from {0}".format(input_config))
    with open(input_config) as icf:
        content = yaml.safe_load(icf)

    if "remotes" in content:
        for remote in content["remotes"]:
            # dict.get() used here to set name=None if no name specified in input cfg file.
            omnicache.AddRemote(remote["url"], name=remote.get("name"))

    return 0


def ScanDirectory(omnicache, scanpath):
    """Recursively scans a directory for git repositories and adds remotes and submodule remotes to omnicache."""
    logging.info("Scanning {0} for remotes to add.".format(scanpath))
    if (not os.path.isdir(scanpath)):
        logging.critical("specified scan path is invalid.")
        return -1

    for (dirpath, dirnames, filenames) in os.walk(scanpath):
        if (".git" in dirnames):
            newRemotes = Omnicache.GetRemotes(dirpath)
            for (name, url) in newRemotes.items():
                omnicache.AddRemote(url, name)

        if (".gitmodules" in filenames):
            out = StringIO()
            ret = RunCmd("git", "config --file .gitmodules --get-regexp url", workingdir=dirpath, outstream=out)
            if (ret == 0):
                for submodule in out.getvalue().splitlines():
                    url = submodule.split()[1]
                    name = submodule.split()[0].split(".")[1]
                    omnicache.AddRemote(url, name)

    return 0


def Export(omnicache, exportPath):
    """Exports omnicache configuration to YAML."""
    logging.info("Exporting omnicache config for {0} to {1}".format(omnicache.path, exportPath))
    content = []
    for (name, data) in omnicache.GetRemoteData().items():
        remoteToWrite = {"url": data["url"]}
        if ("displayname" in data):
            remoteToWrite["name"] = data["displayname"]
        else:
            remoteToWrite["name"] = name
        content.append(remoteToWrite)

    with open(exportPath, "w") as ocf:
        yaml.dump({"remotes": content}, ocf)

    return 0


def get_cli_options():
    """Add CLI arguments to argparse for controlling the omnicache."""
    parser = argparse.ArgumentParser(description='Tool to provide easy method create and manage the OMNICACHE', )
    parser.add_argument(dest="cache_dir", help="path to an existing or desired OMNICACHE directory")
    parser.add_argument("--scan", dest="scan", default=None,
                        help="Scans the path provided for top-level folders with repos to add to the OMNICACHE")
    parser.add_argument("--new", dest="new", help="Initialize the OMNICACHE.  MUST NOT EXIST",
                        action="store_true", default=False)
    parser.add_argument("--init", dest="init", help="Initialize the OMNICACHE if it doesn't already exist",
                        action="store_true", default=False)
    parser.add_argument("-l", "--list", dest="list", default=False, action="store_true",
                        help="List config of OMNICACHE")
    parser.add_argument("-a", "--add", dest="add", nargs=2, action="append",
                        help="Add config entry to OMNICACHE <name> <url>",
                        default=[])
    parser.add_argument("-c", "--configfile", dest="input_config_file", default=None,
                        help="Add new entries from config file to OMNICACHE")
    parser.add_argument("-e", "--exportConfig", dest="output_config_file", default=None,
                        help="Export current omnicache config as a yaml file")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-u", "--update", "--fetch", dest="fetch", action="store_true",
                       help="Update the Omnicache.  All cache changes also cause a fetch", default=False)
    group.add_argument("--no-fetch", dest="no_fetch", action="store_true",
                       help="Prevent auto-fetch if implied by other arguments.", default=False)
    group.add_argument("--fetch-jobs", dest="fetch_jobs", type=int,
                       help="Specify the number of parallel threads (jobs) for fetch operation.", default=0)
    parser.add_argument("-r", "--remove", dest="remove", nargs=1, action="append",
                        help="remove config entry from OMNICACHE <name>", default=[])
    parser.add_argument('--version', action='version', version='%(prog)s ' + OMNICACHE_VERSION)
    parser.add_argument("--debug", dest="debug", help="Output all debug messages to console",
                        action="store_true", default=False)
    args = parser.parse_args()
    return args


def main():
    """Main entry point to managing the omnicache."""
    # setup main console as logger
    logger = logging.getLogger('')
    logger.setLevel(logging.NOTSET)
    console = edk2_logging.setup_console_logging(False)
    logger.addHandler(console)

    args = get_cli_options()

    if args.debug:
        console.setLevel(logging.DEBUG)

    logging.info("Log Started: " + datetime.datetime.strftime(
        datetime.datetime.now(), "%A, %B %d, %Y %I:%M%p"))

    args.cache_dir = os.path.realpath(os.path.abspath(args.cache_dir))
    logging.debug("OMNICACHE dir: {0}".format(args.cache_dir))

    if args.input_config_file is not None:
        args.input_config_file = os.path.realpath(os.path.abspath(args.input_config_file))
        if not os.path.isfile(args.input_config_file):
            logging.critical("Invalid -c argument given.  File ({0}) isn't valid".format(args.input_config_file))
            return -4

    logging.debug("Args: " + str(args))

    omnicache = None
    auto_fetch = False
    # new: initialize omnicache and error if it already exists
    if args.new:
        if os.path.isdir(args.cache_dir):
            logging.critical("--new argument given but OMNICACHE path already exists!")
            return -1
        omnicache = Omnicache(args.cache_dir, create=True)
        auto_fetch = True

    # init: initialize omnicache if not already initialized.
    if args.init and omnicache is None:
        omnicache = Omnicache(args.cache_dir, create=True, convert=True)
        auto_fetch = True

    # Other args require omnicache, so check that it exists.
    if omnicache is None:
        omnicache = Omnicache(args.cache_dir)

    # add: add new source(s) to omnicache from command line arg.
    if (len(args.add) > 0):
        for (name, url) in args.add:
            ret = omnicache.AddRemote(url, name)
            if (ret != 0):
                return -3
        auto_fetch = True

    # config: add or update sources(s) from config file.
    if (args.input_config_file is not None):
        ret = ProcessInputConfig(omnicache, args.input_config_file)
        if (ret != 0):
            return -4
        auto_fetch = True

    # remove: remove source(s) from omnicache as specified by command line arg.
    if (len(args.remove) > 0):
        for url in args.remove:
            ret = omnicache.RemoveRemote(url)
            if (ret != 0):
                return -4

    # scan: recursively scan the given directory and add all repos and submodules
    if (args.scan is not None):
        ret = ScanDirectory(omnicache, args.scan)
        if (ret != 0):
            return -5
        auto_fetch = True

    # fetch: update the omnicache with objects from its remotes.
    # Note: errors are ignored here, since transient network failures may occur that prevent cache update. Those just
    # mean the omnicache may be a a little stale which should not be fatal to users of the cache.
    if (args.fetch or (auto_fetch and not args.no_fetch)):
        omnicache.Fetch(args.fetch_jobs)

    # list: print out the omnicache contents.
    if (args.list):
        omnicache.List()

    # export:
    if (args.output_config_file):
        ret = Export(omnicache, args.output_config_file)
        if (ret != 0):
            return -6

    return 0


if __name__ == '__main__':
    retcode = main()
    logging.shutdown()
    sys.exit(retcode)



##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This file exists to satisfy pythons packaging requirements.

Read more: https://docs.python.org/3/reference/import.html#regular-packages
"""



# @file dsc_processor_plugin
# Plugin for parsing DSCs
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Plugin for parsing DSCs."""


class IDscProcessorPlugin(object):
    """Plugin for parsing DSCs."""

    def do_transform(self, dsc, thebuilder):
        """Does the transform on a DSC.

        Args:
            dsc (obj): the in-memory model of the DSC
            thebuilder (UefiBuilder): UefiBuild object for env information

        Returns:
            (int): 0 or NonZero for success or failure
        """
        return 0

    def get_level(self, thebuilder):
        """Gets the level that this transform operates at.

        Args:
            thebuilder (UefiBuilder): UefiBuild object for env information

        Returns:
            (int): the level

        """
        return 0



# @file web_dependency.py
# This module implements ExternalDependency for files that are available for download online.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An ExternalDependency subclass able to download from a url."""

import logging
import os
import pathlib
import shutil
import tarfile
import tempfile
import urllib.error
import urllib.request
import zipfile

from edk2toollib.utility_functions import RemoveTree

from edk2toolext.environment.external_dependency import ExternalDependency


class WebDependency(ExternalDependency):
    """An ExternalDependency subclass able to download from a url.

    Able to download an uncrompressed, zip, or tar file from the web and can
    optionally compare the hash to a provided hash.

    Attributes:
        internal_path (str): Describes layout of what we're downloading.
                             Include / at the beginning if the ext_dep is a
                             directory. Item located at internal_path will
                             unpacked into the ext_dep folder and this is what
                             the path/shell vars will point to when
                             compute_published_path is run.
        compression_type (str): Supports zip and tar. If the file isn't
                                compressed, do not include this field. Optional
        sha256 (str): Hash of downloaded file to be checked against. Optional

    !!! tip
        The attributes are what must be described in the ext_dep yaml file!
    """

    TypeString = "web"

    def __init__(self, descriptor):
        """Inits a web dependency based off the provided descriptor."""
        super().__init__(descriptor)
        self.internal_path = os.path.normpath(descriptor['internal_path'])
        self.compression_type = descriptor.get('compression_type', None)
        self.sha256 = descriptor.get('sha256', None)

        # If the internal path starts with a / that means we are downloading a directory
        self.download_is_directory = self.internal_path.startswith(os.path.sep)

        # Now we can get rid of the leading /
        self.internal_path = self.internal_path.strip(os.path.sep)

    def __str__(self):
        """Returns a string representation."""
        return f"WebDependecy: {self.source}@{self.version}"

    @staticmethod
    def linuxize_path(path):
        """Replaces windows style separators with linux style separators.

        Arguments:
            path (str): the path
        """
        return "/".join(path.split("\\"))

    @staticmethod
    def unpack(compressed_file_path, destination, internal_path, compression_type):
        """Unpacks a compressed file to the specified location; zip or tar supported.

        Arguments:
            compressed_file_path (str): Name of compressed file to unpack.
            destination (str): directory to unpacked into.
            internal_path (str): internal structure of the compressed volume that you would like extracted.
            compression_type (str): type of compression. tar and zip supported.
        """
        # First, we will open the file depending on the type of compression we're dealing with.

        # tarfile and zipfile both use the Linux path seperator / instead of using os.sep
        linux_internal_path = WebDependency.linuxize_path(internal_path)

        if compression_type == "zip":
            logging.info(f"{compressed_file_path} is a zip file, trying to unpack it.")
            _ref = zipfile.ZipFile(compressed_file_path, 'r')
            files_in_volume = _ref.namelist()

        elif compression_type and "tar" in compression_type:
            logging.info(f"{compressed_file_path} is a tar file, trying to unpack it.")
            # r:* tells tarfile to look at the header and figure out how to extract it
            _ref = tarfile.open(compressed_file_path, "r:*")
            files_in_volume = _ref.getnames()

        else:
            raise RuntimeError(f"{compressed_file_path} was labeled as {compression_type}, which is not supported.")

        # Filter the files inside to only the ones that are inside the important folder
        files_to_extract = [name for name in files_in_volume if linux_internal_path in name]

        for file in files_to_extract:
            _ref.extract(member=file, path=destination)

            # unzip functionality does not preserve file permissions. Fix-up the permissions
            path = pathlib.Path(destination, file)
            if path.is_file() and compression_type == "zip":
                expected_mode = _ref.getinfo(file).external_attr >> 16
                path.chmod(expected_mode)

        _ref.close()

    def fetch(self):
        """Fetches the dependency using internal state from the init."""
        # First, check the global cache to see if it's present.
        if super().fetch():
            return

        url = self.source
        temp_folder = tempfile.mkdtemp()
        temp_file_path = os.path.join(temp_folder, f"{self.name}_{self.version}")

        try:
            # Download the file and save it locally under `temp_file_path`
            with urllib.request.urlopen(url) as response, open(temp_file_path, 'wb') as out_file:
                out_file.write(response.read())
        except urllib.error.HTTPError as e:
            logging.error(f"ran into an issue when resolving ext_dep {self.name} at {self.source}")
            raise e

        # check if file hash is as expected, if it was provided in the ext_dep.json
        if self.sha256:
            with open(temp_file_path, "rb") as file:
                import hashlib
                temp_file_sha256 = hashlib.sha256(file.read()).hexdigest()

            # compare sha256 hexdigests as lowercase to make case insensitive
            if temp_file_sha256.lower() != self.sha256.lower():
                raise RuntimeError(f"{self.name} - sha256 does not match\n\tdownloaded:"
                                   f"\t{temp_file_sha256}\n\tin json:\t{self.sha256}")

        if os.path.isfile(temp_file_path) is False:
            raise RuntimeError(f"{self.name} did not download")

        # Next, we will look at what's inside it and pull out the parts we need.
        if self.compression_type:
            WebDependency.unpack(temp_file_path, temp_folder, self.internal_path, self.compression_type)

        # internal_path points to the "important" part of the ext_dep we're unpacking
        complete_internal_path = os.path.join(temp_folder, self.internal_path)

        # # If we're unpacking a directory, we can copy the important parts into
        # # a directory named self.contents_dir
        if self.download_is_directory:
            logging.info(f"Copying directory from {complete_internal_path} to {self.contents_dir}")
            if os.path.isdir(complete_internal_path) is False:
                # internal_path was not accurate, exit
                raise RuntimeError(f"{self.name} was expecting {complete_internal_path} to exist after unpacking")

            # Move the important folder out and rename it to contents_dir
            shutil.move(complete_internal_path, self.contents_dir)

            # If the unzipped directory still exists, delete it.
            if os.path.isdir(temp_folder):
                logging.debug(f"Cleaning up {temp_folder}")
                RemoveTree(temp_folder)

        # If we just downloaded a file, we need to create a directory named self.contents_dir,
        # copy the file inside, and name it self.internal_path
        else:
            os.makedirs(self.contents_dir, exist_ok=True)
            complete_internal_path = os.path.join(self.contents_dir, self.internal_path)
            logging.info(f"Copying file to {complete_internal_path}")
            shutil.move(temp_file_path, complete_internal_path)

        self.copy_to_global_cache(self.contents_dir)

        # Add a file to track the state of the dependency.
        self.update_state_file()

        # The published path may change now that the package has been unpacked.
        self.published_path = self.compute_published_path()



# @file NugetPublishing.py
# This tool allows a user to create a configuration for nuget as well as
# pack and push (publishing) a release to a feed.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Provides configuration, packing, and publishing nuget packages to a release feed."""
import argparse
import datetime
import logging
import os
import shutil
import sys
import xml.etree.ElementTree as etree
from io import StringIO

import yaml
from edk2toollib.utility_functions import RunCmd

from edk2toolext.environment.extdeptypes.nuget_dependency import NugetDependency

# Use associated SPDX license Identifier
# https://learn.microsoft.com/en-us/nuget/reference/nuspec#license
LICENSE_IDENTIFIER_SUPPORTED = {
    "BSD2": "BSD-2-Clause",
    "BSD3": "BSD-3-Clause",
    "APACHE2": "Apache-2.0",
    "MSPL": "MS-PL",  # Microsoft Public License
    "MIT": "MIT",
    "BSD2pP": "BSD-2-Clause-Patent",  # BSD2 + Patent
}


class NugetSupport(object):
    """Support object for Nuget Publishing tool to configure NuPkg information, pack and send."""
    # NOTE: This *should* have a namespace (http://schemas.microsoft.com/packaging/2010/07/nuspec.xsd)
    #       but ElementTree is incredibly stupid with namespaces.
    NUSPEC_TEMPLATE_XML = r'''<?xml version="1.0" encoding="utf-8"?>
<package>
    <metadata>
        <!-- Required elements-->
        <id></id>
        <version></version>
        <description></description>
        <authors></authors>

        <!-- Optional elements -->
        <license></license>
        <releaseNotes></releaseNotes>
        <projectUrl></projectUrl>
        <repository />
        <copyright></copyright>
        <tags></tags>
    </metadata>
    <!-- Optional 'files' node -->
    <files>
        <file src="" target="" />
    </files>
</package>'''

    RELEASE_NOTE_SHORT_STRING_MAX_LENGTH = 500

    def __init__(self, Name=None, ConfigFile=None):
        """Inits a new NugetSupport object.

        for new instances without existing config provide the Name parameter.
        for creating instance based on config file provide the path to the ConfigFile
        """
        self.Name = Name
        self.TempFileToDelete = []  # everytime a temp is created add to list to cleanup
        self.NewVersion = None
        self.ConfigChanged = False

        if (ConfigFile is not None):
            self.FromConfigfile(ConfigFile)
            self.Name = self.ConfigData["name"]
        else:
            if (Name is None):
                raise ValueError("Cannot construct object with both Name and ConfigFile as None")
            self.ConfigData = {"name": Name}
            self.Config = None

    def CleanUp(self):
        """Remove all temporary files."""
        logging.debug("CleanUp Called.  Deleting all Temp Files")
        for a in self.TempFileToDelete:
            os.remove(a)

    def ToConfigFile(self, filepath=None):
        """Save config to a yaml file."""
        if (not self.ConfigChanged):
            logging.debug("No Config Changes.  Skip Writing config file")
            return 0

        if (filepath is None and self.Config is None):
            logging.error("No Config File to save to.")
            return -1

        if (filepath is not None):
            self.Config = filepath

        if (filepath is None):
            logging.error("No filepath for Config File")

        with open(filepath, "w") as c:
            yaml.dump(self.ConfigData, c, indent=4)
        logging.debug("Wrote config file to: %s" % filepath)
        self.ConfigChanged = False
        return 0

    def FromConfigfile(self, filepath):
        """Load config from a yaml file."""
        self.Config = filepath
        with open(self.Config, "r") as c:
            self.ConfigData = yaml.safe_load(c)

    def SetBasicData(self, authors, license, project, description, server, copyright,
                     repositoryType=None, repositoryUrl=None, repositoryBranch=None,
                     repositoryCommit=None):
        """Set basic data in the config data."""
        self.ConfigData["author_string"] = authors
        if license:
            self.ConfigData["license"] = license
        self.ConfigData["project_url"] = project
        self.ConfigData["description_string"] = description
        self.ConfigData["server_url"] = server
        if repositoryType:
            self.ConfigData["repository_type"] = repositoryType
        if repositoryUrl:
            self.ConfigData["repository_url"] = repositoryUrl
        if repositoryBranch:
            self.ConfigData["repository_branch"] = repositoryBranch
        if repositoryCommit:
            self.ConfigData["repository_commit"] = repositoryCommit
        if not copyright:
            copyright = "Copyright %d" % datetime.date.today().year
        self.ConfigData["copyright_string"] = copyright

        self.ConfigData["tags_string"] = ""

        self.ConfigChanged = True

    def UpdateLicensePath(self, licensepath):
        """Update license in the config data.

        Update license in the config data with an absolute path to a license
        to pack.
        """
        self.ConfigData["license"] = licensepath

    def IsValidLicense(self):
        """Returns whether the License is valid."""
        if "license" not in self.ConfigData:
            return False

        license = self.ConfigData["license"]

        if license in LICENSE_IDENTIFIER_SUPPORTED.values():
            return True

        if not os.path.isabs(license) or not os.path.isfile(license):
            return False

        if os.path.basename(license).lower() not in ["license.txt", "license.md"]:
            return False

        return True

    def UpdateCopyright(self, copyright):
        """Update copyright in the config data."""
        self.ConfigData["copyright_string"] = copyright
        self.ConfigChanged = True

    def UpdateTags(self, tags=[]):
        """Update tags in the config data."""
        self.ConfigData["tags_string"] = " ".join(tags)
        self.ConfigChanged = True

    def UpdateRepositoryInfo(self, type=None, url=None, branch=None, commit=None):
        """Update repository information."""
        if type:
            self.ConfigData["repository_type"] = type
            self.ConfigChanged = True
        if url:
            self.ConfigData["repository_url"] = url
            self.ConfigChanged = True
        if branch:
            self.ConfigData["repository_branch"] = branch
            self.ConfigChanged = True
        if commit:
            self.ConfigData["repository_commit"] = commit
            self.ConfigChanged = True

    def Print(self):
        """Print info about the Nuget Object."""
        print("=======================================")
        print(" Name:        " + self.Name)
        if (self.Config):
            print(" ConfigFile:  " + self.Config)
        else:
            print(" ConfigFile:  NOT SET")

        for k, v in self.ConfigData.items():
            print(" %s:   %s" % (k, v))

        print("----------------------------------------")
        print(" Temp Files List: ")
        for a in self.TempFileToDelete:
            print("    " + a)
        print("-----------------------------------------")
        print("=======================================")

    def LogObject(self):
        """Logs info about Nuget Object to the logger."""
        logging.debug("=======================================")
        logging.debug(" Name:        " + self.Name)
        if (self.Config):
            logging.debug(" ConfigFile:  " + self.Config)
        else:
            logging.debug(" ConfigFile:  NOT SET")

        for k, v in self.ConfigData.items():
            logging.debug(" %s:   %s" % (k, v))

        logging.debug("----------------------------------------")
        logging.debug(" Temp Files List: ")
        for a in self.TempFileToDelete:
            logging.debug("    " + a)
        logging.debug("-----------------------------------------")
        logging.debug("=======================================")

    #
    # given NugetSupport object
    # create a nuspec file for packing
    #

    def _MakeNuspecXml(self, ContentDir, ReleaseNotesText=None):
        package = etree.fromstring(NugetSupport.NUSPEC_TEMPLATE_XML)
        package.attrib["xmlns"] = "http://schemas.microsoft.com/packaging/2010/07/nuspec.xsd"
        meta = package.find("./metadata")
        meta.find("id").text = self.Name
        meta.find("version").text = self.NewVersion
        meta.find("authors").text = self.ConfigData["author_string"]
        meta.find("projectUrl").text = self.ConfigData["project_url"]
        repository_item_present = bool([k for k in self.ConfigData.keys() if "repository" in k.lower()])
        r = meta.find("repository")
        if repository_item_present:
            if "repository_type" in self.ConfigData:
                r.set("type", self.ConfigData["repository_type"])
            if "repository_url" in self.ConfigData:
                r.set("url", self.ConfigData["repository_url"])
            if "repository_branch" in self.ConfigData:
                r.set("branch", self.ConfigData["repository_branch"])
            if "repository_commit" in self.ConfigData:
                r.set("commit", self.ConfigData["repository_commit"])
        else:
            meta.remove(r)
        meta.find("description").text = self.ConfigData["description_string"]
        meta.find("copyright").text = self.ConfigData["copyright_string"]
        if "tags_string" in self.ConfigData:
            meta.find("tags").text = self.ConfigData["tags_string"]
        files = package.find("files")
        f = files.find("file")
        f.set("target", self.Name)
        f.set("src", ContentDir + "\\**\\*")

        # Add appropriate attribute to the license element depending on if the
        # license is custom or not. If the license is custom, add a new file
        # element pointing at the license
        if os.path.isfile(self.ConfigData["license"]):
            meta.find("license").text = os.path.basename(self.ConfigData["license"])
            meta.find("license").attrib["type"] = "file"
            f = etree.Element("file", attrib={"src": self.ConfigData["license"], "target": ""})
            files.append(f)
        else:
            meta.find("license").text = self.ConfigData["license"]
            meta.find("license").attrib["type"] = "expression"

        if (ReleaseNotesText is not None):
            logging.debug("Make Nuspec Xml - ReleaseNotesText is not none.")
            #
            # Make sure it doesn't exceed reasonable length of string
            #
            if (len(ReleaseNotesText) > NugetSupport.RELEASE_NOTE_SHORT_STRING_MAX_LENGTH):
                logging.info("Make Nuspec Xml - ReleaseNotesText too long.  Length is (%d)" % len(ReleaseNotesText))
                logging.debug("Original ReleaseNotesText is: %s" % ReleaseNotesText)
                # cut it off at max length
                ReleaseNotesText = ReleaseNotesText[:NugetSupport.RELEASE_NOTE_SHORT_STRING_MAX_LENGTH]
                # walk back to trim at last end of sentence
                ReleaseNotesText = ReleaseNotesText.rpartition(".")[0].strip()
                logging.debug("New ReleaseNotesText is: %s" % ReleaseNotesText)

            meta.find("releaseNotes").text = ReleaseNotesText
        else:
            logging.debug("Make Nuspec Xml - ReleaseNotesText None. Removing element from nuspec.")
            meta.remove(meta.find("releaseNotes"))

        return etree.tostring(package)

    def _GetNuPkgFileName(self, version):
        # Nuget removes leading zeros so to match we must do the same
        s = self.Name + "."
        append_tag = None
        parts = version.split(".")
        if "-" in parts[-1]:
            parts[-1], append_tag = parts[-1].split("-")
        int_parts = [str(int(a)) for a in parts]

        # nuget must have at least x.y.z and will make zero any element undefined
        for _ in range(len(int_parts), 3):
            int_parts.append("0")
        # Join the integers together
        s += ".".join(int_parts)
        if append_tag is not None:
            s += f"-{append_tag}"
        s += ".nupkg"
        return s

    def Pack(self, version, OutputDirectory, ContentDir, RelNotesText=None):
        """Pack the current contents into Nupkg."""
        self.NewVersion = version

        # content must be absolute path in nuspec otherwise it is assumed
        # relative to nuspec file.
        cdir = os.path.abspath(ContentDir)

        # make nuspec file
        xmlstring = self._MakeNuspecXml(cdir, RelNotesText)
        nuspec = os.path.join(OutputDirectory, self.Name + ".nuspec")
        self.TempFileToDelete.append(nuspec)
        f = open(nuspec, "wb")
        f.write(xmlstring)
        f.close()

        # run nuget
        cmd = NugetDependency.GetNugetCmd()
        cmd += ["pack", nuspec]
        cmd += ["-OutputDirectory", '"' + OutputDirectory + '"']
        cmd += ["-Verbosity", "detailed"]
        # cmd += ["-NonInteractive"]
        ret = RunCmd(cmd[0], " ".join(cmd[1:]))

        if (ret != 0):
            logging.error("Failed on nuget command.  RC = 0x%x" % ret)
            return ret

        self.NuPackageFile = os.path.join(OutputDirectory, self._GetNuPkgFileName(self.NewVersion))
        self.TempFileToDelete.append(self.NuPackageFile)
        return ret

    def Push(self, nuPackage, apikey):
        """Push nuget package to the server.

        Raises:
            (Exception): file path is invalid
        """
        if (not os.path.isfile(nuPackage)):
            raise Exception("Invalid file path for NuPkg file")
        logging.debug("Pushing %s file to server %s" % (nuPackage, self.ConfigData["server_url"]))

        cmd = NugetDependency.GetNugetCmd()
        cmd += ["push", nuPackage]
        cmd += ["-Verbosity", "detailed"]
        # cmd += ["-NonInteractive"]
        cmd += ["-Source", self.ConfigData["server_url"]]
        cmd += ["-ApiKey", apikey]
        output_buffer = StringIO()
        ret = RunCmd(cmd[0], " ".join(cmd[1:]), outstream=output_buffer)

        if (ret != 0):
            # Rewind the buffer and capture the contents.
            output_buffer.seek(0)
            output_contents = output_buffer.read()

            # Check for the API message.
            if "API key is invalid".lower() in output_contents.lower():
                logging.critical("API key is invalid. Please use --ApiKey to provide a valid key.")

            # Generic error.
            logging.error("Failed on nuget command.  RC = 0x%x" % ret)

        return ret


def GatherArguments():
    """Adds CLI arguments for controlling the nuget_publishing tool."""
    tempparser = argparse.ArgumentParser(
        description='Nuget Helper Script for creating, packing, and pushing packages', add_help=False)
    tempparser.add_argument('--Operation', dest="op", choices=["New", "Pack", "Push", "PackAndPush"], required=True)

    # Get the operation the user wants to do
    (args, rest) = tempparser.parse_known_args()

    # now build up the real parser with required parameters
    parser = argparse.ArgumentParser(description='Nuget Helper Script for creating, packing, and pushing packages')
    parser.add_argument("--Dirty", dest="Dirty", action="store_true", help="Keep all temp files", default=False)
    parser.add_argument('--Operation', dest="Operation", choices=["New", "Pack", "Push", "PackAndPush"], required=True)
    parser.add_argument("--OutputLog", dest="OutputLog", help="Create an output log file")

    if (args.op.lower() == "new"):
        parser.add_argument("--ConfigFileFolderPath", dest="ConfigFileFolderPath",
                            help="<Required>Path to folder to save new config file to", required=True)
        parser.add_argument('--Name',
                            dest='Name',
                            help='<Required> The unique id/name of the package.  This is a string naming the package',
                            required=True)
        parser.add_argument('--Author', dest="Author", help="<Required> Author string for publishing", required=True)
        parser.add_argument("--ProjectUrl", dest="Project", help="<Required> Project Url", required=True)
        repo_group = parser.add_argument_group(title="Repository Parameters",
                                               description="Optional Repository Parameters")
        repo_group.add_argument("--RepositoryType", dest="RepositoryType", help="<Optional> Repository Type",
                                required=False)
        repo_group.add_argument("--RepositoryUrl", dest="RepositoryUrl", help="<Optional> Repository Url",
                                required=False)
        repo_group.add_argument("--RepositoryBranch", dest="RepositoryBranch", help="<Optional> Repository Branch",
                                required=False)
        repo_group.add_argument("--RepositoryCommit", dest="RepositoryCommit", help="<Optional> Repository Commit",
                                required=False)
        parser.add_argument('--LicenseIdentifier', dest="LicenseIdentifier", default=None,
                            choices=LICENSE_IDENTIFIER_SUPPORTED.keys(), help="Standard Licenses")
        parser.add_argument('--Description', dest="Description",
                            help="<Required> Description of package.", required=True)
        parser.add_argument("--FeedUrl", dest="FeedUrl",
                            help="<Required>Feed Url of the nuget server feed", required=True)
        parser.add_argument('--Copyright', dest="Copyright", help="Copyright string", required=False)

    elif (args.op.lower() == "pack" or args.op.lower() == "packandpush"):
        parser.add_argument("--ConfigFilePath", dest="ConfigFilePath",
                            help="<Required>Path to config file", required=True)
        parser.add_argument('--Version', dest="Version", help="<Required> Version to publish", required=True)
        parser.add_argument('--ReleaseNotesText', dest="ReleaseNotes",
                            help="<Optional>Release Notes String", required=False)
        parser.add_argument('--InputFolderPath', dest="InputFolderPath",
                            help="<Required>Relative/Absolute Path to folder containing content to pack.",
                            required=True)
        parser.add_argument('--Copyright', dest="Copyright", help="<Optional>Change the Copyright string")
        parser.add_argument('--t', "-tag", dest="Tags", type=str,
                            help="<Optional>Add tags to the nuspec. Multiple are --t Tag1,Tag2 or --t Tag1 --t Tag2",
                            action="append", default=[])
        parser.add_argument('--ApiKey', dest="ApiKey",
                            help="<Optional>Api key to use. Default is 'VSTS' which will invoke interactive login",
                            default="VSTS")
        parser.add_argument('--CustomLicensePath', dest="CustomLicensePath", default=None,
                            help="<Optional> If CustomLicense set in `new` phase, provide absolute path of License \
                            File to pack. Does not override existing valid license.")
        repo_group = parser.add_argument_group(title="Repository Parameters",
                                               description="Optional Repository Parameters")
        repo_group.add_argument("--RepositoryType", dest="RepositoryType", help="<Optional> Repository Type",
                                required=False)
        repo_group.add_argument("--RepositoryUrl", dest="RepositoryUrl", help="<Optional> Change the repository Url",
                                required=False)
        repo_group.add_argument("--RepositoryBranch", dest="RepositoryBranch",
                                help="<Optional> Change the repository branch", required=False)
        repo_group.add_argument("--RepositoryCommit", dest="RepositoryCommit",
                                help="<Optional> Change the repository commit", required=False)

    elif (args.op.lower() == "push"):
        parser.add_argument("--ConfigFilePath", dest="ConfigFilePath",
                            help="<Required>Path to config file",
                            required=True)
        parser.add_argument('--PackageFile', dest="PackageFile", help="<Required>Path To Package File", required=True)
        parser.add_argument('--ApiKey', dest="ApiKey",
                            help="<Optional>Api key to use. Default is 'VSTS' which will invoke interactive login",
                            default="VSTS")

    if (args.op.lower() == "pack"):
        parser.add_argument('--OutputFolderPath',
                            dest="OutputFolderPath",
                            help="<Optional>Output folder where nupkg will be saved.  Default is cwd",
                            default=os.getcwd())

    return parser.parse_args()


def main():
    """Entry point into nuget_publishing after initial configuration."""
    args = GatherArguments()
    ret = 0

    # setup file based logging if outputReport specified
    if (args.OutputLog):
        if (len(args.OutputLog) < 2):
            logging.critical("the output log file parameter is invalid")
            return -2

        # setup file based logging
        filelogger = logging.FileHandler(filename=args.OutputLog, mode='w')
        filelogger.setLevel(logging.DEBUG)
        logging.getLogger('').addHandler(filelogger)

    logging.info("Log Started: " + datetime.datetime.strftime(datetime.datetime.now(), "%A, %B %d, %Y %I:%M%p"))

    TempOutDir = None
    NuPkgFilePath = None

    if (args.Operation.lower() == "new"):
        logging.critical("Generating new nuget configuration...")
        logging.debug("Checking input parameters for new")
        ConfigFilePath = os.path.join(args.ConfigFileFolderPath, args.Name.strip() + ".config.yaml")

        if (not os.path.isdir(args.ConfigFileFolderPath)):
            logging.critical("Config File Folder Path doesn't exist.  %s" % args.ConfigFileFolderPath)
            raise Exception("Invalid Config File Folder.  Doesn't exist")

        if (os.path.isfile(ConfigFilePath)):
            logging.critical("Config File already exists at that path.  %s" % ConfigFilePath)
            raise Exception("Can't Create New Config file when file already exists")

        nu = NugetSupport(Name=args.Name)

        # Provide Standard License Identifier in New Command
        # Or provide Custom License Path in Pack Command.
        if args.LicenseIdentifier is None:
            license = None
        else:
            license = LICENSE_IDENTIFIER_SUPPORTED[args.LicenseIdentifier]

        nu.SetBasicData(
            args.Author,
            license,
            args.Project,
            args.Description,
            args.FeedUrl,
            args.Copyright,
            args.RepositoryType,
            args.RepositoryUrl,
            args.RepositoryBranch,
            args.RepositoryCommit)
        nu.LogObject()
        ret = nu.ToConfigFile(ConfigFilePath)
        return ret

    elif (args.Operation.lower() == "pack" or args.Operation.lower() == "packandpush"):
        logging.critical("Creating nuget package")
        logging.debug("Checking input parameters for packing")
        # check args
        if (not os.path.isfile(args.ConfigFilePath)):
            logging.critical("Invalid Config File (%s).  File doesn't exist" % args.ConfigFilePath)
            raise Exception("Invalid Config File.  File doesn't exist")
        if (not os.path.isdir(args.InputFolderPath)):
            logging.critical("Invalid Input folder (%s).  Folder doesn't exist" % args.InputFolderPath)
            raise Exception("Invalid Input folder.  folder doesn't exist")
        contents = os.listdir(args.InputFolderPath)
        logging.debug("Input Folder contains %d files" % len(contents))
        if (len(contents) == 0):
            logging.critical("No binary contents to pack in %s" % args.InputFolderPath)
            raise Exception("No binary contents to package")

        # make a temp dir for the pack operation which actually creates files
        TempOutDir = os.path.join(os.getcwd(), "_TEMP_" + str(datetime.datetime.now().time()).replace(":", "_"))
        os.mkdir(TempOutDir)

        nu = NugetSupport(ConfigFile=args.ConfigFilePath)

        if not nu.IsValidLicense():
            # Invalid License and not setting it with a custom License
            if args.CustomLicensePath is None:
                logging.critical("Standard License not found in config file and custom license not provided.")
                logging.critical("Provide a custom license path with --CustomLicensePath.")
                raise Exception("Invalid License.")
            nu.UpdateLicensePath(args.CustomLicensePath)

        if not nu.IsValidLicense():
            logging.critical("Invalid Custom License")
            logging.critical("    Verify custom license file name is license.txt or license.md")
            logging.critical("    Verify custom license file path is in absolute format and valid")
            raise Exception("Invalid License.")

        if (args.Copyright is not None):
            nu.UpdateCopyright(args.Copyright)

        nu.UpdateRepositoryInfo(args.RepositoryType, args.RepositoryUrl,
                                args.RepositoryBranch, args.RepositoryCommit)

        if (len(args.Tags) > 0):
            tagListSet = set()
            for item in args.Tags:  # Parse out the individual packages
                item_list = item.split(",")
                for individual_item in item_list:
                    # in case cmd line caller used Windows folder slashes
                    individual_item = individual_item.replace("\\", "/")
                    tagListSet.add(individual_item.strip())
            tagList = list(tagListSet)
            nu.UpdateTags(tagList)
        '''
        ret = nu.ToConfigFile()
        if (ret != 0):
            logging.error("Failed to save config file.  Return Code 0x%x" % ret)
            return ret
        '''

        ret = nu.Pack(args.Version, TempOutDir, args.InputFolderPath, args.ReleaseNotes)
        if (ret != 0):
            logging.error("Failed to pack.  Return Code 0x%x" % ret)
            return ret

        NuPkgFilePath = nu.NuPackageFile

    if (args.Operation.lower() == "pack"):
        if (not os.path.isdir(args.OutputFolderPath)):
            logging.critical("Invalid Pack Output Folder (%s).  Folder doesn't exist" % args.OutputFolderPath)
            raise Exception("Invalid Output folder.  folder doesn't exist")
        # since it is pack only lets copy nupkg file to output
        shutil.copyfile(NuPkgFilePath, os.path.join(args.OutputFolderPath, os.path.basename(NuPkgFilePath)))
        NuPkgFilePath = os.path.join(args.OutputFolderPath, os.path.basename(NuPkgFilePath))

    if (args.Operation.lower() == "push"):
        # set the parameters for push
        logging.debug("Checking input parameters for push")
        # check args
        if (not os.path.isfile(args.ConfigFilePath)):
            logging.critical("Invalid Config File (%s).  File doesn't exist" % args.ConfigFilePath)
            raise Exception("Invalid Config File.  File doesn't exist")
        NuPkgFilePath = args.PackageFile
        nu = NugetSupport(ConfigFile=args.ConfigFilePath)

    if (args.Operation.lower() == "push" or args.Operation.lower() == "packandpush"):
        # do the pushing
        logging.critical("Pushing the package")
        logging.debug("NuPkgFilePath is %s" % NuPkgFilePath)
        # check args
        if (not os.path.isfile(NuPkgFilePath)):
            logging.critical("NuPkgFilePath is not valid file.  %s" % NuPkgFilePath)
            raise Exception("Invalid Pkg File.  File doesn't exist")
        ret = nu.Push(NuPkgFilePath, args.ApiKey)

    nu.LogObject()
    nu.ToConfigFile(args.ConfigFilePath)  # save any changes
    if (not args.Dirty):
        nu.CleanUp()
        if (TempOutDir is not None):
            os.removedirs(TempOutDir)
    return ret


def go():
    """Main entry into the nuget publishing tool."""
    # setup main console as logger
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(levelname)s - %(message)s")
    console = logging.StreamHandler()
    console.setLevel(logging.WARNING)
    console.setFormatter(formatter)
    logger.addHandler(console)

    # call main worker function
    retcode = main()

    if retcode != 0:
        logging.critical("Failed.  Return Code: %d" % retcode)
    else:
        logging.critical("Success!")
    # end logging
    logging.shutdown()
    sys.exit(retcode)


if __name__ == '__main__':
    go()



# @file GitDependency.py
# This module implements ExternalDependency for a git repository
# This should only be used for read-only repositories. Any changes in
# these extdeps will be removed.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An ExternalDependency subclass able to clone from git."""
import logging
import os
from urllib.parse import urlsplit, urlunsplit

from edk2toolext.environment import repo_resolver, shell_environment
from edk2toolext.environment.external_dependency import ExternalDependency


class GitDependency(ExternalDependency):
    """An ExternalDependency subclass able to clone from git.

    Attributes:
        source (str): url for git clone
        version (str): commit from git repo
        url_creds_var (str): shell_var name for credential updating. Optional

    !!! tip
        The attributes are what must be described in the ext_dep yaml file!
    """
    TypeString = "git"

    def __init__(self, descriptor):
        """Inits a git dependency based off the provided descriptor."""
        super().__init__(descriptor)

        # Check to see whether this URL should be patched.
        url_creds_var = descriptor.get('url_creds_var', None)
        if url_creds_var is not None:
            env = shell_environment.GetEnvironment()
            url_creds = env.get_shell_var(url_creds_var)
            if url_creds is not None:
                # Break things up.
                source_parts = urlsplit(self.source)
                # Modify the URL host with the creds.
                new_parts = (source_parts.scheme,
                             url_creds + '@' + source_parts.netloc,
                             source_parts.path,
                             source_parts.query,
                             source_parts.fragment)
                # Put things back together.
                self.source = urlunsplit(new_parts)

        self.repo_url = self.source
        self.commit = self.version
        self._local_repo_root_path = os.path.join(os.path.abspath(self.contents_dir), self.name)
        self.logger = logging.getLogger("git-dependency")

        # valid_attributes = ["Path", "Url", "Branch", "Commit", "ReferencePath", "Full"]
        self._repo_resolver_dep_obj = {"Path": self.name, "Url": self.repo_url, "Commit": self.commit}

    def __str__(self):
        """Return a string representation."""
        return f"GitDependecy: {self.repo_url}@{self.commit}"

    def fetch(self):
        """Fetches the dependency using internal state from the init."""
        try:
            repo_resolver.resolve(self._local_repo_root_path, self._repo_resolver_dep_obj, update_ok=True)
        except repo_resolver.GitCommandError as e:
            logging.debug(f'Cmd failed for git dependency: {self._local_repo_root_path}')
            logging.debug(e)

        # Add a file to track the state of the dependency.
        self.update_state_file()

    def clean(self):
        """Removes the local clone of the repo."""
        self.logger.debug("Cleaning git dependency directory for '%s'..." % self.name)

        if os.path.isdir(self._local_repo_root_path):
            # Clean up git dependency specific stuff
            repo_resolver.clear_folder(self.contents_dir)

        # Let super class clean up common dependency stuff
        super().clean()

    def verify(self):
        """Verifies the clone was successful."""
        result = True
        details = repo_resolver.repo_details(self._local_repo_root_path)

        if not details['Path'].is_dir():
            self.logger.info('Not a directory')
            result = False

        elif not any(details['Path'].iterdir()):
            self.logger.info('No files in directory')
            result = False

        elif not details['Initialized']:
            self.logger.info('Not Initialized')
            result = False

        elif details['Dirty']:
            self.logger.info('Dirty')
            result = False

        elif self.version.lower() not in [details['Head']['HexSha'], details['Head']['HexShaShort']]:
            self.logger.info(f'Mismatched sha: [head: {details["Head"]["HexSha"]}], [expected: {self.version}]')
            result = False

        self.logger.debug("Verify '%s' returning '%s'." % (self.name, result))
        return result

    def compute_published_path(self):
        """Override to include the repository name in the published path."""
        return os.path.join(super().compute_published_path(), self.name)
